# SPDX-FileCopyrightText: Copyright (c) 2025-2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: LicenseRef-NVIDIA-SOFTWARE-LICENSE

# CUDA Toolkit v13.1.1
_FALLBACK_EXPLANATIONS = {
    0: (
        "The API call returned with no errors. In the case of query calls, this"
        " also means that the operation being queried is complete (see"
        " ::cuEventQuery() and ::cuStreamQuery())."
    ),
    1: (
        "This indicates that one or more of the parameters passed to the API call"
        " is not within an acceptable range of values."
    ),
    2: (
        "The API call failed because it was unable to allocate enough memory or"
        " other resources to perform the requested operation."
    ),
    3: (
        "This indicates that the CUDA driver has not been initialized with"
        " ::cuInit() or that initialization has failed."
    ),
    4: "This indicates that the CUDA driver is in the process of shutting down.",
    5: (
        "This indicates profiler is not initialized for this run. This can"
        " happen when the application is running with external profiling tools"
        " like visual profiler."
    ),
    6: (
        "This error return is deprecated as of CUDA 5.0. It is no longer an error"
        " to attempt to enable/disable the profiling via ::cuProfilerStart or"
        " ::cuProfilerStop without initialization."
    ),
    7: (
        "This error return is deprecated as of CUDA 5.0. It is no longer an error"
        " to call cuProfilerStart() when profiling is already enabled."
    ),
    8: (
        "This error return is deprecated as of CUDA 5.0. It is no longer an error"
        " to call cuProfilerStop() when profiling is already disabled."
    ),
    34: (
        "This indicates that the CUDA driver that the application has loaded is a"
        " stub library. Applications that run with the stub rather than a real"
        " driver loaded will result in CUDA API returning this error."
    ),
    36: (
        "This indicates that the API call requires a newer CUDA driver than the one"
        " currently installed. Users should install an updated NVIDIA CUDA driver"
        " to allow the API call to succeed."
    ),
    46: (
        "This indicates that requested CUDA device is unavailable at the current"
        " time. Devices are often unavailable due to use of"
        " ::CU_COMPUTEMODE_EXCLUSIVE_PROCESS or ::CU_COMPUTEMODE_PROHIBITED."
    ),
    100: ("This indicates that no CUDA-capable devices were detected by the installed CUDA driver."),
    101: (
        "This indicates that the device ordinal supplied by the user does not"
        " correspond to a valid CUDA device or that the action requested is"
        " invalid for the specified device."
    ),
    102: "This error indicates that the Grid license is not applied.",
    200: ("This indicates that the device kernel image is invalid. This can also indicate an invalid CUDA module."),
    201: (
        "This most frequently indicates that there is no context bound to the"
        " current thread. This can also be returned if the context passed to an"
        " API call is not a valid handle (such as a context that has had"
        " ::cuCtxDestroy() invoked on it). This can also be returned if a user"
        " mixes different API versions (i.e. 3010 context with 3020 API calls)."
        " See ::cuCtxGetApiVersion() for more details."
        " This can also be returned if the green context passed to an API call"
        " was not converted to a ::CUcontext using ::cuCtxFromGreenCtx API."
    ),
    202: (
        "This indicated that the context being supplied as a parameter to the"
        " API call was already the active context."
        " This error return is deprecated as of CUDA 3.2. It is no longer an"
        " error to attempt to push the active context via ::cuCtxPushCurrent()."
    ),
    205: "This indicates that a map or register operation has failed.",
    206: "This indicates that an unmap or unregister operation has failed.",
    207: ("This indicates that the specified array is currently mapped and thus cannot be destroyed."),
    208: "This indicates that the resource is already mapped.",
    209: (
        "This indicates that there is no kernel image available that is suitable"
        " for the device. This can occur when a user specifies code generation"
        " options for a particular CUDA source file that do not include the"
        " corresponding device configuration."
    ),
    210: "This indicates that a resource has already been acquired.",
    211: "This indicates that a resource is not mapped.",
    212: ("This indicates that a mapped resource is not available for access as an array."),
    213: ("This indicates that a mapped resource is not available for access as a pointer."),
    214: ("This indicates that an uncorrectable ECC error was detected during execution."),
    215: ("This indicates that the ::CUlimit passed to the API call is not supported by the active device."),
    216: (
        "This indicates that the ::CUcontext passed to the API call can"
        " only be bound to a single CPU thread at a time but is already"
        " bound to a CPU thread."
    ),
    217: ("This indicates that peer access is not supported across the given devices."),
    218: "This indicates that a PTX JIT compilation failed.",
    219: "This indicates an error with OpenGL or DirectX context.",
    220: ("This indicates that an uncorrectable NVLink error was detected during the execution."),
    221: "This indicates that the PTX JIT compiler library was not found.",
    222: "This indicates that the provided PTX was compiled with an unsupported toolchain.",
    223: "This indicates that the PTX JIT compilation was disabled.",
    224: ("This indicates that the ::CUexecAffinityType passed to the API call is not supported by the active device."),
    225: (
        "This indicates that the code to be compiled by the PTX JIT contains unsupported call to cudaDeviceSynchronize."
    ),
    226: (
        "This indicates that an exception occurred on the device that is now"
        " contained by the GPU's error containment capability. Common causes are -"
        " a. Certain types of invalid accesses of peer GPU memory over nvlink"
        " b. Certain classes of hardware errors"
        " This leaves the process in an inconsistent state and any further CUDA"
        " work will return the same error. To continue using CUDA, the process must"
        " be terminated and relaunched."
    ),
    300: (
        "This indicates that the device kernel source is invalid. This includes"
        " compilation/linker errors encountered in device code or user error."
    ),
    301: "This indicates that the file specified was not found.",
    302: "This indicates that a link to a shared object failed to resolve.",
    303: "This indicates that initialization of a shared object failed.",
    304: "This indicates that an OS call failed.",
    400: (
        "This indicates that a resource handle passed to the API call was not"
        " valid. Resource handles are opaque types like ::CUstream and ::CUevent."
    ),
    401: (
        "This indicates that a resource required by the API call is not in a"
        " valid state to perform the requested operation."
    ),
    402: (
        "This indicates an attempt was made to introspect an object in a way that"
        " would discard semantically important information. This is either due to"
        " the object using funtionality newer than the API version used to"
        " introspect it or omission of optional return arguments."
    ),
    500: (
        "This indicates that a named symbol was not found. Examples of symbols"
        " are global/constant variable names, driver function names, texture names,"
        " and surface names."
    ),
    600: (
        "This indicates that asynchronous operations issued previously have not"
        " completed yet. This result is not actually an error, but must be indicated"
        " differently than ::CUDA_SUCCESS (which indicates completion). Calls that"
        " may return this value include ::cuEventQuery() and ::cuStreamQuery()."
    ),
    700: (
        "While executing a kernel, the device encountered a"
        " load or store instruction on an invalid memory address."
        " This leaves the process in an inconsistent state and any further CUDA work"
        " will return the same error. To continue using CUDA, the process must be terminated"
        " and relaunched."
    ),
    701: (
        "This indicates that a launch did not occur because it did not have"
        " appropriate resources. This error usually indicates that the user has"
        " attempted to pass too many arguments to the device kernel, or the"
        " kernel launch specifies too many threads for the kernel's register"
        " count. Passing arguments of the wrong size (i.e. a 64-bit pointer"
        " when a 32-bit int is expected) is equivalent to passing too many"
        " arguments and can also result in this error."
    ),
    702: (
        "This indicates that the device kernel took too long to execute. This can"
        " only occur if timeouts are enabled - see the device attribute"
        " ::CU_DEVICE_ATTRIBUTE_KERNEL_EXEC_TIMEOUT for more information."
        " This leaves the process in an inconsistent state and any further CUDA work"
        " will return the same error. To continue using CUDA, the process must be terminated"
        " and relaunched."
    ),
    703: ("This error indicates a kernel launch that uses an incompatible texturing mode."),
    704: (
        "This error indicates that a call to ::cuCtxEnablePeerAccess() is"
        " trying to re-enable peer access to a context which has already"
        " had peer access to it enabled."
    ),
    705: (
        "This error indicates that ::cuCtxDisablePeerAccess() is"
        " trying to disable peer access which has not been enabled yet"
        " via ::cuCtxEnablePeerAccess()."
    ),
    708: ("This error indicates that the primary context for the specified device has already been initialized."),
    709: (
        "This error indicates that the context current to the calling thread"
        " has been destroyed using ::cuCtxDestroy, or is a primary context which"
        " has not yet been initialized."
    ),
    710: (
        "A device-side assert triggered during kernel execution. The context"
        " cannot be used anymore, and must be destroyed. All existing device"
        " memory allocations from this context are invalid and must be"
        " reconstructed if the program is to continue using CUDA."
    ),
    711: (
        "This error indicates that the hardware resources required to enable"
        " peer access have been exhausted for one or more of the devices"
        " passed to ::cuCtxEnablePeerAccess()."
    ),
    712: ("This error indicates that the memory range passed to ::cuMemHostRegister() has already been registered."),
    713: (
        "This error indicates that the pointer passed to ::cuMemHostUnregister()"
        " does not correspond to any currently registered memory region."
    ),
    714: (
        "While executing a kernel, the device encountered a stack error."
        " This can be due to stack corruption or exceeding the stack size limit."
        " This leaves the process in an inconsistent state and any further CUDA work"
        " will return the same error. To continue using CUDA, the process must be terminated"
        " and relaunched."
    ),
    715: (
        "While executing a kernel, the device encountered an illegal instruction."
        " This leaves the process in an inconsistent state and any further CUDA work"
        " will return the same error. To continue using CUDA, the process must be terminated"
        " and relaunched."
    ),
    716: (
        "While executing a kernel, the device encountered a load or store instruction"
        " on a memory address which is not aligned."
        " This leaves the process in an inconsistent state and any further CUDA work"
        " will return the same error. To continue using CUDA, the process must be terminated"
        " and relaunched."
    ),
    717: (
        "While executing a kernel, the device encountered an instruction"
        " which can only operate on memory locations in certain address spaces"
        " (global, shared, or local), but was supplied a memory address not"
        " belonging to an allowed address space."
        " This leaves the process in an inconsistent state and any further CUDA work"
        " will return the same error. To continue using CUDA, the process must be terminated"
        " and relaunched."
    ),
    718: (
        "While executing a kernel, the device program counter wrapped its address space."
        " This leaves the process in an inconsistent state and any further CUDA work"
        " will return the same error. To continue using CUDA, the process must be terminated"
        " and relaunched."
    ),
    719: (
        "An exception occurred on the device while executing a kernel. Common"
        " causes include dereferencing an invalid device pointer and accessing"
        " out of bounds shared memory. Less common cases can be system specific - more"
        " information about these cases can be found in the system specific user guide."
        " This leaves the process in an inconsistent state and any further CUDA work"
        " will return the same error. To continue using CUDA, the process must be terminated"
        " and relaunched."
    ),
    720: (
        "This error indicates that the number of blocks launched per grid for a kernel that was"
        " launched via either ::cuLaunchCooperativeKernel or ::cuLaunchCooperativeKernelMultiDevice"
        " exceeds the maximum number of blocks as allowed by ::cuOccupancyMaxActiveBlocksPerMultiprocessor"
        " or ::cuOccupancyMaxActiveBlocksPerMultiprocessorWithFlags times the number of multiprocessors"
        " as specified by the device attribute ::CU_DEVICE_ATTRIBUTE_MULTIPROCESSOR_COUNT."
    ),
    721: (
        "An exception occurred on the device while exiting a kernel using tensor memory: the"
        " tensor memory was not completely deallocated. This leaves the process in an inconsistent"
        " state and any further CUDA work will return the same error. To continue using CUDA, the"
        " process must be terminated and relaunched."
    ),
    800: "This error indicates that the attempted operation is not permitted.",
    801: ("This error indicates that the attempted operation is not supported on the current system or device."),
    802: (
        "This error indicates that the system is not yet ready to start any CUDA"
        " work.  To continue using CUDA, verify the system configuration is in a"
        " valid state and all required driver daemons are actively running."
        " More information about this error can be found in the system specific"
        " user guide."
    ),
    803: (
        "This error indicates that there is a mismatch between the versions of"
        " the display driver and the CUDA driver. Refer to the compatibility documentation"
        " for supported versions."
    ),
    804: (
        "This error indicates that the system was upgraded to run with forward compatibility"
        " but the visible hardware detected by CUDA does not support this configuration."
        " Refer to the compatibility documentation for the supported hardware matrix or ensure"
        " that only supported hardware is visible during initialization via the CUDA_VISIBLE_DEVICES"
        " environment variable."
    ),
    805: "This error indicates that the MPS client failed to connect to the MPS control daemon or the MPS server.",
    806: "This error indicates that the remote procedural call between the MPS server and the MPS client failed.",
    807: (
        "This error indicates that the MPS server is not ready to accept new MPS client requests."
        " This error can be returned when the MPS server is in the process of recovering from a fatal failure."
    ),
    808: "This error indicates that the hardware resources required to create MPS client have been exhausted.",
    809: "This error indicates the the hardware resources required to support device connections have been exhausted.",
    810: "This error indicates that the MPS client has been terminated by the server. To continue using CUDA, the process must be terminated and relaunched.",
    811: "This error indicates that the module is using CUDA Dynamic Parallelism, but the current configuration, like MPS, does not support it.",
    812: "This error indicates that a module contains an unsupported interaction between different versions of CUDA Dynamic Parallelism.",
    900: ("This error indicates that the operation is not permitted when the stream is capturing."),
    901: (
        "This error indicates that the current capture sequence on the stream"
        " has been invalidated due to a previous error."
    ),
    902: (
        "This error indicates that the operation would have resulted in a merge of two independent capture sequences."
    ),
    903: "This error indicates that the capture was not initiated in this stream.",
    904: ("This error indicates that the capture sequence contains a fork that was not joined to the primary stream."),
    905: (
        "This error indicates that a dependency would have been created which"
        " crosses the capture sequence boundary. Only implicit in-stream ordering"
        " dependencies are allowed to cross the boundary."
    ),
    906: ("This error indicates a disallowed implicit dependency on a current capture sequence from cudaStreamLegacy."),
    907: (
        "This error indicates that the operation is not permitted on an event which"
        " was last recorded in a capturing stream."
    ),
    908: (
        "A stream capture sequence not initiated with the ::CU_STREAM_CAPTURE_MODE_RELAXED"
        " argument to ::cuStreamBeginCapture was passed to ::cuStreamEndCapture in a"
        " different thread."
    ),
    909: "This error indicates that the timeout specified for the wait operation has lapsed.",
    910: (
        "This error indicates that the graph update was not performed because it included"
        " changes which violated constraints specific to instantiated graph update."
    ),
    911: (
        "This indicates that an async error has occurred in a device outside of CUDA."
        " If CUDA was waiting for an external device's signal before consuming shared data,"
        " the external device signaled an error indicating that the data is not valid for"
        " consumption. This leaves the process in an inconsistent state and any further CUDA"
        " work will return the same error. To continue using CUDA, the process must be"
        " terminated and relaunched."
    ),
    912: "Indicates a kernel launch error due to cluster misconfiguration.",
    913: ("Indiciates a function handle is not loaded when calling an API that requires a loaded function."),
    914: ("This error indicates one or more resources passed in are not valid resource types for the operation."),
    915: ("This error indicates one or more resources are insufficient or non-applicable for the operation."),
    916: ("This error indicates that an error happened during the key rotation sequence."),
    917: (
        "This error indicates that the requested operation is not permitted because the"
        " stream is in a detached state. This can occur if the green context associated"
        " with the stream has been destroyed, limiting the stream's operational capabilities."
    ),
    999: "This indicates that an unknown internal error has occurred.",
}



# SPDX-FileCopyrightText: Copyright (c) 2024-2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
#
# SPDX-License-Identifier: Apache-2.0

from libc.stddef cimport size_t
from libc.stdint cimport intptr_t

from libcpp.memory cimport shared_ptr

from cuda.bindings cimport cydriver
from cuda.bindings cimport cynvrtc
from cuda.bindings cimport cynvvm
from cuda.bindings cimport cynvjitlink


# =============================================================================
# Handle type aliases and inline helpers (declared from C++ header)
# =============================================================================

cdef extern from "_cpp/resource_handles.hpp" namespace "cuda_core":
    # Handle types
    ctypedef shared_ptr[const cydriver.CUcontext] ContextHandle
    ctypedef shared_ptr[const cydriver.CUstream] StreamHandle
    ctypedef shared_ptr[const cydriver.CUevent] EventHandle
    ctypedef shared_ptr[const cydriver.CUmemoryPool] MemoryPoolHandle
    ctypedef shared_ptr[const cydriver.CUdeviceptr] DevicePtrHandle
    ctypedef shared_ptr[const cydriver.CUlibrary] LibraryHandle
    ctypedef shared_ptr[const cydriver.CUkernel] KernelHandle
    ctypedef shared_ptr[const cydriver.CUgraph] GraphHandle
    ctypedef shared_ptr[const cydriver.CUgraphNode] GraphNodeHandle
    ctypedef shared_ptr[const cydriver.CUgraphicsResource] GraphicsResourceHandle
    ctypedef shared_ptr[const cynvrtc.nvrtcProgram] NvrtcProgramHandle

    # NvvmProgramValue and NvJitLinkValue are TaggedHandle<void*, Tag>
    # instantiations that make each shared_ptr type distinct for overloading.
    cppclass NvvmProgramValue "cuda_core::NvvmProgramValue":
        pass
    cppclass NvJitLinkValue "cuda_core::NvJitLinkValue":
        pass
    ctypedef shared_ptr[const NvvmProgramValue] NvvmProgramHandle
    ctypedef shared_ptr[const NvJitLinkValue] NvJitLinkHandle

    ctypedef shared_ptr[const cydriver.CUlinkState] CuLinkHandle
    ctypedef shared_ptr[const int] FileDescriptorHandle

    # as_cu() - extract the raw CUDA handle (inline C++)
    cydriver.CUcontext as_cu(ContextHandle h) noexcept nogil
    cydriver.CUstream as_cu(StreamHandle h) noexcept nogil
    cydriver.CUevent as_cu(EventHandle h) noexcept nogil
    cydriver.CUmemoryPool as_cu(MemoryPoolHandle h) noexcept nogil
    cydriver.CUdeviceptr as_cu(DevicePtrHandle h) noexcept nogil
    cydriver.CUlibrary as_cu(LibraryHandle h) noexcept nogil
    cydriver.CUkernel as_cu(KernelHandle h) noexcept nogil
    cydriver.CUgraph as_cu(GraphHandle h) noexcept nogil
    cydriver.CUgraphNode as_cu(GraphNodeHandle h) noexcept nogil
    cydriver.CUgraphicsResource as_cu(GraphicsResourceHandle h) noexcept nogil
    cynvrtc.nvrtcProgram as_cu(NvrtcProgramHandle h) noexcept nogil
    cynvvm.nvvmProgram as_cu(NvvmProgramHandle h) noexcept nogil
    cynvjitlink.nvJitLinkHandle as_cu(NvJitLinkHandle h) noexcept nogil
    cydriver.CUlinkState as_cu(CuLinkHandle h) noexcept nogil

    # as_intptr() - extract handle as intptr_t for Python interop (inline C++)
    intptr_t as_intptr(ContextHandle h) noexcept nogil
    intptr_t as_intptr(StreamHandle h) noexcept nogil
    intptr_t as_intptr(EventHandle h) noexcept nogil
    intptr_t as_intptr(MemoryPoolHandle h) noexcept nogil
    intptr_t as_intptr(DevicePtrHandle h) noexcept nogil
    intptr_t as_intptr(LibraryHandle h) noexcept nogil
    intptr_t as_intptr(KernelHandle h) noexcept nogil
    intptr_t as_intptr(GraphHandle h) noexcept nogil
    intptr_t as_intptr(GraphNodeHandle h) noexcept nogil
    intptr_t as_intptr(GraphicsResourceHandle h) noexcept nogil
    intptr_t as_intptr(NvrtcProgramHandle h) noexcept nogil
    intptr_t as_intptr(NvvmProgramHandle h) noexcept nogil
    intptr_t as_intptr(NvJitLinkHandle h) noexcept nogil
    intptr_t as_intptr(CuLinkHandle h) noexcept nogil
    intptr_t as_intptr(FileDescriptorHandle h) noexcept nogil

    # as_py() - convert handle to Python wrapper object (inline C++; requires GIL)
    object as_py(ContextHandle h)
    object as_py(StreamHandle h)
    object as_py(EventHandle h)
    object as_py(MemoryPoolHandle h)
    object as_py(DevicePtrHandle h)
    object as_py(LibraryHandle h)
    object as_py(KernelHandle h)
    object as_py(GraphHandle h)
    object as_py(GraphNodeHandle h)
    object as_py(GraphicsResourceHandle h)
    object as_py(NvrtcProgramHandle h)
    object as_py(NvvmProgramHandle h)
    object as_py(NvJitLinkHandle h)
    object as_py(CuLinkHandle h)
    object as_py(FileDescriptorHandle h)


# =============================================================================
# Wrapper function declarations (implemented in _resource_handles.pyx)
#
# Consumer modules cimport these. Calls go through _resource_handles.so.
# =============================================================================

# Thread-local error handling
cdef cydriver.CUresult get_last_error() noexcept nogil
cdef cydriver.CUresult peek_last_error() noexcept nogil
cdef void clear_last_error() noexcept nogil

# Context handles
cdef ContextHandle create_context_handle_ref(cydriver.CUcontext ctx) except+ nogil
cdef ContextHandle get_primary_context(int device_id) except+ nogil
cdef ContextHandle get_current_context() except+ nogil

# Stream handles
cdef StreamHandle create_stream_handle(
    const ContextHandle& h_ctx, unsigned int flags, int priority) except+ nogil
cdef StreamHandle create_stream_handle_ref(cydriver.CUstream stream) except+ nogil
cdef StreamHandle create_stream_handle_with_owner(cydriver.CUstream stream, object owner) except+ nogil
cdef StreamHandle get_legacy_stream() except+ nogil
cdef StreamHandle get_per_thread_stream() except+ nogil

# Event handles
cdef EventHandle create_event_handle(
    const ContextHandle& h_ctx, unsigned int flags,
    bint timing_disabled, bint busy_waited,
    bint ipc_enabled, int device_id) except+ nogil
cdef EventHandle create_event_handle_noctx(unsigned int flags) except+ nogil
cdef EventHandle create_event_handle_ref(cydriver.CUevent event) except+ nogil
cdef EventHandle create_event_handle_ipc(
    const cydriver.CUipcEventHandle& ipc_handle, bint busy_waited) except+ nogil

# Event metadata getters
cdef bint get_event_timing_disabled(const EventHandle& h) noexcept nogil
cdef bint get_event_busy_waited(const EventHandle& h) noexcept nogil
cdef bint get_event_ipc_enabled(const EventHandle& h) noexcept nogil
cdef int get_event_device_id(const EventHandle& h) noexcept nogil
cdef ContextHandle get_event_context(const EventHandle& h) noexcept nogil

# Memory pool handles
cdef MemoryPoolHandle create_mempool_handle(
    const cydriver.CUmemPoolProps& props) except+ nogil
cdef MemoryPoolHandle create_mempool_handle_ref(cydriver.CUmemoryPool pool) except+ nogil
cdef MemoryPoolHandle get_device_mempool(int device_id) except+ nogil
cdef MemoryPoolHandle create_mempool_handle_ipc(
    int fd, cydriver.CUmemAllocationHandleType handle_type) except+ nogil

# Device pointer handles
cdef DevicePtrHandle deviceptr_alloc_from_pool(
    size_t size, const MemoryPoolHandle& h_pool, const StreamHandle& h_stream) except+ nogil
cdef DevicePtrHandle deviceptr_alloc_async(size_t size, const StreamHandle& h_stream) except+ nogil
cdef DevicePtrHandle deviceptr_alloc(size_t size) except+ nogil
cdef DevicePtrHandle deviceptr_alloc_host(size_t size) except+ nogil
cdef DevicePtrHandle deviceptr_create_ref(cydriver.CUdeviceptr ptr) except+ nogil
cdef DevicePtrHandle deviceptr_create_with_owner(cydriver.CUdeviceptr ptr, object owner) except+ nogil
cdef DevicePtrHandle deviceptr_create_mapped_graphics(
    cydriver.CUdeviceptr ptr,
    const GraphicsResourceHandle& h_resource,
    const StreamHandle& h_stream) except+ nogil
cdef DevicePtrHandle deviceptr_create_with_mr(
    cydriver.CUdeviceptr ptr, size_t size, object mr) except+ nogil

# MR deallocation callback type and registration
ctypedef void (*MRDeallocCallback)(
    object mr, cydriver.CUdeviceptr ptr, size_t size,
    const StreamHandle& stream) noexcept
cdef void register_mr_dealloc_callback(MRDeallocCallback cb) noexcept

cdef DevicePtrHandle deviceptr_import_ipc(
    const MemoryPoolHandle& h_pool, const void* export_data, const StreamHandle& h_stream) except+ nogil
cdef StreamHandle deallocation_stream(const DevicePtrHandle& h) noexcept nogil
cdef void set_deallocation_stream(const DevicePtrHandle& h, const StreamHandle& h_stream) noexcept nogil

# Library handles
cdef LibraryHandle create_library_handle_from_file(const char* path) except+ nogil
cdef LibraryHandle create_library_handle_from_data(const void* data) except+ nogil
cdef LibraryHandle create_library_handle_ref(cydriver.CUlibrary library) except+ nogil

# Kernel handles
cdef KernelHandle create_kernel_handle(const LibraryHandle& h_library, const char* name) except+ nogil
cdef KernelHandle create_kernel_handle_ref(cydriver.CUkernel kernel) except+ nogil
cdef LibraryHandle get_kernel_library(const KernelHandle& h) noexcept nogil

# Graph handles
cdef GraphHandle create_graph_handle(cydriver.CUgraph graph) except+ nogil
cdef GraphHandle create_graph_handle_ref(cydriver.CUgraph graph, const GraphHandle& h_parent) except+ nogil

# Graph node handles
cdef GraphNodeHandle create_graph_node_handle(cydriver.CUgraphNode node, const GraphHandle& h_graph) except+ nogil
cdef GraphHandle graph_node_get_graph(const GraphNodeHandle& h) noexcept nogil
cdef void invalidate_graph_node(const GraphNodeHandle& h) noexcept nogil

# Graphics resource handles
cdef GraphicsResourceHandle create_graphics_resource_handle(
    cydriver.CUgraphicsResource resource) except+ nogil

# NVRTC Program handles
cdef NvrtcProgramHandle create_nvrtc_program_handle(cynvrtc.nvrtcProgram prog) except+ nogil
cdef NvrtcProgramHandle create_nvrtc_program_handle_ref(cynvrtc.nvrtcProgram prog) except+ nogil

# NVVM Program handles
cdef NvvmProgramHandle create_nvvm_program_handle(cynvvm.nvvmProgram prog) except+ nogil
cdef NvvmProgramHandle create_nvvm_program_handle_ref(cynvvm.nvvmProgram prog) except+ nogil

# nvJitLink handles
cdef NvJitLinkHandle create_nvjitlink_handle(cynvjitlink.nvJitLinkHandle handle) except+ nogil
cdef NvJitLinkHandle create_nvjitlink_handle_ref(cynvjitlink.nvJitLinkHandle handle) except+ nogil

# cuLink handles
cdef CuLinkHandle create_culink_handle(cydriver.CUlinkState state) except+ nogil
cdef CuLinkHandle create_culink_handle_ref(cydriver.CUlinkState state) except+ nogil

# File descriptor handles
cdef FileDescriptorHandle create_fd_handle(int fd) except+ nogil
cdef FileDescriptorHandle create_fd_handle_ref(int fd) except+ nogil



# SPDX-FileCopyrightText: Copyright (c) 2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
#
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from collections.abc import Callable, Iterable
from dataclasses import dataclass


@dataclass(frozen=True)
class PeerAccessPlan:
    """Normalized peer-access target state and the driver updates it requires."""

    target_ids: tuple[int, ...]
    to_add: tuple[int, ...]
    to_remove: tuple[int, ...]


def normalize_peer_access_targets(
    owner_device_id: int,
    requested_devices: Iterable[object],
    *,
    resolve_device_id: Callable[[object], int],
) -> tuple[int, ...]:
    """Return sorted, unique peer device IDs, excluding the owner device."""

    target_ids = {resolve_device_id(device) for device in requested_devices}
    target_ids.discard(owner_device_id)
    return tuple(sorted(target_ids))


def plan_peer_access_update(
    owner_device_id: int,
    current_peer_ids: Iterable[int],
    requested_devices: Iterable[object],
    *,
    resolve_device_id: Callable[[object], int],
    can_access_peer: Callable[[int], bool],
) -> PeerAccessPlan:
    """Compute the peer-access target state and add/remove deltas."""

    target_ids = normalize_peer_access_targets(
        owner_device_id,
        requested_devices,
        resolve_device_id=resolve_device_id,
    )
    bad = tuple(dev_id for dev_id in target_ids if not can_access_peer(dev_id))
    if bad:
        bad_ids = ", ".join(str(dev_id) for dev_id in bad)
        raise ValueError(f"Device {owner_device_id} cannot access peer(s): {bad_ids}")

    current_ids = set(current_peer_ids)
    target_id_set = set(target_ids)
    return PeerAccessPlan(
        target_ids=target_ids,
        to_add=tuple(sorted(target_id_set - current_ids)),
        to_remove=tuple(sorted(current_ids - target_id_set)),
    )



# SPDX-FileCopyrightText: Copyright (c) 2024-2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
#
# SPDX-License-Identifier: Apache-2.0

"""
Backward compatibility stubs for cuda.core.experimental namespace.

This module provides forwarding stubs that import from the new cuda.core.*
locations and emit deprecation warnings. Users should migrate to importing
directly from cuda.core instead of cuda.core.experimental.

The experimental namespace will be removed in v1.0.0.

"""


def _warn_deprecated():
    """Emit a deprecation warning for using the experimental namespace.

    Note: This warning is only when the experimental module is first imported.
    Subsequent accesses to attributes (like utils, Device, etc.) do not trigger
    additional warnings since they are already set in the module namespace.
    """
    import warnings

    warnings.warn(
        "The cuda.core.experimental namespace is deprecated. "
        "Please import directly from cuda.core instead. "
        "For example, use 'from cuda.core import Device' instead of "
        "'from cuda.core.experimental import Device'. "
        "The experimental namespace will be removed in v1.0.0.",
        DeprecationWarning,
        stacklevel=3,
    )


# Import from new locations and re-export
_warn_deprecated()


from cuda.core import system, utils

# Make utils accessible as a submodule for backward compatibility
__import__("sys").modules[__spec__.name + ".utils"] = utils


from cuda.core._device import Device
from cuda.core._event import Event, EventOptions
from cuda.core._launch_config import LaunchConfig
from cuda.core._launcher import launch
from cuda.core._layout import _StridedLayout
from cuda.core._linker import Linker, LinkerOptions
from cuda.core._memory import (
    Buffer,
    DeviceMemoryResource,
    DeviceMemoryResourceOptions,
    GraphMemoryResource,
    LegacyPinnedMemoryResource,
    ManagedMemoryResource,
    ManagedMemoryResourceOptions,
    MemoryResource,
    PinnedMemoryResource,
    PinnedMemoryResourceOptions,
    VirtualMemoryResource,
    VirtualMemoryResourceOptions,
)
from cuda.core._module import Kernel, ObjectCode
from cuda.core._program import Program, ProgramOptions
from cuda.core._stream import Stream, StreamOptions
from cuda.core.graph import (
    Graph,
    GraphBuilder,
    GraphCompleteOptions,
    GraphDebugPrintOptions,
)



# SPDX-FileCopyrightText: Copyright (c) 2024-2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
#
# SPDX-License-Identifier: Apache-2.0

from libc.stdint cimport uintptr_t

from cuda.core._resource_handles cimport DevicePtrHandle
from cuda.core._stream cimport Stream


cdef struct _MemAttrs:
    int device_id
    bint is_device_accessible
    bint is_host_accessible
    bint is_managed


cdef class Buffer:
    cdef:
        DevicePtrHandle _h_ptr
        size_t          _size
        MemoryResource  _memory_resource
        object          _ipc_data
        object          _owner
        _MemAttrs       _mem_attrs
        bint            _mem_attrs_inited
        object          __weakref__


cdef class MemoryResource:
    pass


# Helper function to create a Buffer from a DevicePtrHandle
cdef Buffer Buffer_from_deviceptr_handle(
    DevicePtrHandle h_ptr,
    size_t size,
    MemoryResource mr,
    object ipc_descriptor = *
)



# SPDX-FileCopyrightText: Copyright (c) 2024-2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
#
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from cuda.core._memory._buffer import DevicePointerT

from cuda.core._memory._buffer import Buffer, MemoryResource
from cuda.core._utils.cuda_utils import (
    _check_driver_error as raise_if_driver_error,
)
from cuda.core._utils.cuda_utils import (
    driver,
)

__all__ = ["LegacyPinnedMemoryResource", "_SynchronousMemoryResource"]


class LegacyPinnedMemoryResource(MemoryResource):
    """Create a pinned memory resource that uses legacy cuMemAllocHost/cudaMallocHost
    APIs.
    """

    # TODO: support creating this MR with flags that are later passed to cuMemHostAlloc?

    def allocate(self, size, stream=None) -> Buffer:
        """Allocate a buffer of the requested size.

        Parameters
        ----------
        size : int
            The size of the buffer to allocate, in bytes.
        stream : Stream, optional
            Currently ignored

        Returns
        -------
        Buffer
            The allocated buffer object, which is accessible on both host and device.
        """
        if stream is None:
            from cuda.core._stream import default_stream

            stream = default_stream()
        if size:
            err, ptr = driver.cuMemAllocHost(size)
            raise_if_driver_error(err)
        else:
            ptr = 0
        return Buffer._init(ptr, size, self)

    def deallocate(self, ptr: DevicePointerT, size, stream):
        """Deallocate a buffer previously allocated by this resource.

        Parameters
        ----------
        ptr : :obj:`~_memory.DevicePointerT`
            The pointer or handle to the buffer to deallocate.
        size : int
            The size of the buffer to deallocate, in bytes.
        stream : Stream
            The stream on which to perform the deallocation synchronously.
        """
        if stream is not None:
            stream.sync()

        if size:
            (err,) = driver.cuMemFreeHost(ptr)
            raise_if_driver_error(err)

    @property
    def is_device_accessible(self) -> bool:
        """bool: this memory resource provides device-accessible buffers."""
        return True

    @property
    def is_host_accessible(self) -> bool:
        """bool: this memory resource provides host-accessible buffers."""
        return True

    @property
    def device_id(self) -> int:
        """This memory resource is not bound to any GPU."""
        raise RuntimeError("a pinned memory resource is not bound to any GPU")


class _SynchronousMemoryResource(MemoryResource):
    __slots__ = ("_device_id",)

    def __init__(self, device_id):
        from .._device import Device

        self._device_id = Device(device_id).device_id

    def allocate(self, size, stream=None) -> Buffer:
        if stream is None:
            from cuda.core._stream import default_stream

            stream = default_stream()
        if size:
            err, ptr = driver.cuMemAlloc(size)
            raise_if_driver_error(err)
        else:
            ptr = 0
        return Buffer._init(ptr, size, self)

    def deallocate(self, ptr, size, stream):
        if stream is not None:
            stream.sync()
        if size:
            (err,) = driver.cuMemFree(ptr)
            raise_if_driver_error(err)

    @property
    def is_device_accessible(self) -> bool:
        return True

    @property
    def is_host_accessible(self) -> bool:
        return False

    @property
    def device_id(self) -> int:
        return self._device_id



# SPDX-FileCopyrightText: Copyright (c) 2024-2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
#
# SPDX-License-Identifier: Apache-2.0

from cuda.core._memoryview import (
    StridedMemoryView,  # noqa: F401
    args_viewable_as_strided_memory,  # noqa: F401
)



# SPDX-FileCopyrightText: Copyright (c) 2024-2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
#
# SPDX-License-Identifier: Apache-2.0

cimport cpython

from libc cimport stdlib
from libc.stdint cimport uint8_t
from libc.stdint cimport uint16_t
from libc.stdint cimport uint32_t
from libc.stdint cimport int32_t
from libc.stdint cimport int64_t
from libc.stdint cimport uint64_t
from libc.stdint cimport intptr_t


cdef extern from "_include/dlpack.h" nogil:
    """
    #define DLPACK_TENSOR_UNUSED_NAME "dltensor"
    #define DLPACK_VERSIONED_TENSOR_UNUSED_NAME "dltensor_versioned"
    #define DLPACK_TENSOR_USED_NAME "used_dltensor"
    #define DLPACK_VERSIONED_TENSOR_USED_NAME "used_dltensor_versioned"
    """
    ctypedef enum _DLDeviceType "DLDeviceType":
        _kDLCPU "kDLCPU"
        _kDLCUDA "kDLCUDA"
        _kDLCUDAHost "kDLCUDAHost"
        _kDLCUDAManaged "kDLCUDAManaged"
        _kDLTrn "kDLTrn"

    ctypedef struct DLDevice:
        _DLDeviceType device_type
        int32_t device_id

    cdef enum DLDataTypeCode:
        kDLInt
        kDLUInt
        kDLFloat
        kDLBfloat
        kDLComplex
        kDLBool

    ctypedef struct DLDataType:
        uint8_t code
        uint8_t bits
        uint16_t lanes

    ctypedef struct DLTensor:
        void* data
        DLDevice device
        int32_t ndim
        DLDataType dtype
        int64_t* shape
        int64_t* strides
        uint64_t byte_offset

    ctypedef struct DLManagedTensor:
        DLTensor dl_tensor
        void* manager_ctx
        void (*deleter)(DLManagedTensor*)

    ctypedef struct DLPackVersion:
        uint32_t major
        uint32_t minor

    ctypedef struct DLManagedTensorVersioned:
        DLPackVersion version
        void* manager_ctx
        void (*deleter)(DLManagedTensorVersioned*)
        uint64_t flags
        DLTensor dl_tensor

    int DLPACK_MAJOR_VERSION
    int DLPACK_MINOR_VERSION
    int DLPACK_FLAG_BITMASK_READ_ONLY
    int DLPACK_FLAG_BITMASK_IS_COPIED
    int DLPACK_FLAG_BITMASK_IS_SUBBYTE_TYPE_PADDED

    const char* DLPACK_TENSOR_UNUSED_NAME
    const char* DLPACK_VERSIONED_TENSOR_UNUSED_NAME
    const char* DLPACK_TENSOR_USED_NAME
    const char* DLPACK_VERSIONED_TENSOR_USED_NAME


cdef extern from "_include/dlpack.h":
    ctypedef int (*DLPackManagedTensorAllocator)(
        DLTensor* prototype,
        DLManagedTensorVersioned** out,
        void* error_ctx,
        void (*SetError)(void* error_ctx, const char* kind, const char* message) noexcept
    )

    ctypedef int (*DLPackManagedTensorFromPyObjectNoSync)(
        void* py_object,
        DLManagedTensorVersioned** out
    )

    ctypedef int (*DLPackManagedTensorToPyObjectNoSync)(
        DLManagedTensorVersioned* tensor,
        void** out_py_object
    )

    ctypedef int (*DLPackDLTensorFromPyObjectNoSync)(
        void* py_object,
        DLTensor* out
    )

    ctypedef int (*DLPackCurrentWorkStream)(
        _DLDeviceType device_type,
        int32_t device_id,
        void** out_current_stream
    )

    ctypedef struct DLPackExchangeAPIHeader:
        DLPackVersion version
        DLPackExchangeAPIHeader* prev_api

    ctypedef struct DLPackExchangeAPI:
        DLPackExchangeAPIHeader header
        DLPackManagedTensorAllocator managed_tensor_allocator
        DLPackManagedTensorFromPyObjectNoSync managed_tensor_from_py_object_no_sync
        DLPackManagedTensorToPyObjectNoSync managed_tensor_to_py_object_no_sync
        DLPackDLTensorFromPyObjectNoSync dltensor_from_py_object_no_sync
        DLPackCurrentWorkStream current_work_stream



# SPDX-FileCopyrightText: Copyright (c) 2025-2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: LicenseRef-NVIDIA-SOFTWARE-LICENSE

from cuda.bindings import driver
from cuda.core._utils.enum_explanations_helpers import get_best_available_explanations


def _load_fallback_explanations():
    from cuda.core._utils.driver_cu_result_explanations_frozen import _FALLBACK_EXPLANATIONS

    return _FALLBACK_EXPLANATIONS


DRIVER_CU_RESULT_EXPLANATIONS = get_best_available_explanations(driver.CUresult, _load_fallback_explanations)



# SPDX-FileCopyrightText: Copyright (c) 2024-2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
#
# SPDX-License-Identifier: Apache-2.0

from cuda.core._memory._memory_pool cimport _MemPool
from cuda.core._memory._ipc cimport IPCDataForMR


cdef class PinnedMemoryResource(_MemPool):
    cdef int _numa_id



# SPDX-FileCopyrightText: Copyright (c) 2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
#
# SPDX-License-Identifier: Apache-2.0

"""Public type aliases and protocols used in cuda.core API signatures."""

from cuda.core._memory._buffer import DevicePointerT
from cuda.core._stream import IsStreamT

__all__ = [
    "DevicePointerT",
    "IsStreamT",
]



# SPDX-FileCopyrightText: Copyright (c) 2024-2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
#
# SPDX-License-Identifier: Apache-2.0

from cuda.bindings cimport cydriver
from libc.stdint cimport intptr_t
from cuda.core._memoryview cimport StridedMemoryView


cdef class TensorMapDescriptor:
    cdef cydriver.CUtensorMap _tensor_map
    cdef int _device_id
    cdef intptr_t _context
    cdef object _source_ref
    cdef StridedMemoryView _view_ref
    cdef object _repr_info

    cdef int _check_context_compat(self) except -1
    cdef void* _get_data_ptr(self)



# SPDX-FileCopyrightText: Copyright (c) 2024-2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
#
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Iterable, Literal

if TYPE_CHECKING:
    from cuda.core._stream import Stream

from cuda.core._device import Device
from cuda.core._memory._buffer import Buffer, MemoryResource
from cuda.core._utils.cuda_utils import (
    Transaction,
    check_or_create_options,
    driver,
)
from cuda.core._utils.cuda_utils import (
    _check_driver_error as raise_if_driver_error,
)
from cuda.core._utils.version import binding_version

__all__ = ["VirtualMemoryResource", "VirtualMemoryResourceOptions"]

VirtualMemoryHandleTypeT = Literal["posix_fd", "generic", "win32_kmt", "fabric"] | None
VirtualMemoryLocationTypeT = Literal["device", "host", "host_numa", "host_numa_current"]
VirtualMemoryGranularityT = Literal["minimum", "recommended"]
VirtualMemoryAccessTypeT = Literal["rw", "r"] | None
VirtualMemoryAllocationTypeT = Literal["pinned", "managed"]


@dataclass
class VirtualMemoryResourceOptions:
    """A configuration object for the VirtualMemoryResource
       Stores configuration information which tells the resource how to use the CUDA VMM APIs

    Attributes
    ----------
    allocation_type: :obj:`~_memory.VirtualMemoryAllocationTypeT`
        Controls the type of allocation.
    location_type: :obj:`~_memory.VirtualMemoryLocationTypeT`
        Controls the location of the allocation.
    handle_type: :obj:`~_memory.VirtualMemoryHandleTypeT`
        Export handle type for the physical allocation. Use
        ``"posix_fd"`` on Linux if you plan to
        import/export the allocation (required for cuMemRetainAllocationHandle).
        Use `None` if you don't need an exportable handle.
    gpu_direct_rdma: bool
        Hint that the allocation should be GDR-capable (if supported).
    granularity: :obj:`~_memory.VirtualMemoryGranularityT`
        Controls granularity query and size rounding.
    addr_hint: int
        A (optional) virtual address hint to try to reserve at. Setting it to 0 lets the CUDA driver decide.
    addr_align: int
        Alignment for the VA reservation. If `None`, use the queried granularity.
    peers: Iterable[int]
        Extra device IDs that should be granted access in addition to ``device``.
    self_access: :obj:`~_memory.VirtualMemoryAccessTypeT`
        Access flags for the owning device.
    peer_access: :obj:`~_memory.VirtualMemoryAccessTypeT`
        Access flags for peers.
    """

    # Human-friendly strings; normalized in __post_init__
    allocation_type: VirtualMemoryAllocationTypeT = "pinned"
    location_type: VirtualMemoryLocationTypeT = "device"
    handle_type: VirtualMemoryHandleTypeT = "posix_fd"
    granularity: VirtualMemoryGranularityT = "recommended"
    gpu_direct_rdma: bool = False
    addr_hint: int | None = 0
    addr_align: int | None = None
    peers: Iterable[int] = field(default_factory=tuple)
    self_access: VirtualMemoryAccessTypeT = "rw"
    peer_access: VirtualMemoryAccessTypeT = "rw"

    _a = driver.CUmemAccess_flags
    _access_flags = {"rw": _a.CU_MEM_ACCESS_FLAGS_PROT_READWRITE, "r": _a.CU_MEM_ACCESS_FLAGS_PROT_READ, None: 0}  # noqa: RUF012
    _h = driver.CUmemAllocationHandleType
    _handle_types = {  # noqa: RUF012
        None: _h.CU_MEM_HANDLE_TYPE_NONE,
        "posix_fd": _h.CU_MEM_HANDLE_TYPE_POSIX_FILE_DESCRIPTOR,
        "win32_kmt": _h.CU_MEM_HANDLE_TYPE_WIN32_KMT,
        "fabric": _h.CU_MEM_HANDLE_TYPE_FABRIC,
    }
    _g = driver.CUmemAllocationGranularity_flags
    _granularity = {  # noqa: RUF012
        "recommended": _g.CU_MEM_ALLOC_GRANULARITY_RECOMMENDED,
        "minimum": _g.CU_MEM_ALLOC_GRANULARITY_MINIMUM,
    }
    _l = driver.CUmemLocationType
    _location_type = {  # noqa: RUF012
        "device": _l.CU_MEM_LOCATION_TYPE_DEVICE,
        "host": _l.CU_MEM_LOCATION_TYPE_HOST,
        "host_numa": _l.CU_MEM_LOCATION_TYPE_HOST_NUMA,
        "host_numa_current": _l.CU_MEM_LOCATION_TYPE_HOST_NUMA_CURRENT,
    }
    _t = driver.CUmemAllocationType
    # CUDA 13+ exposes MANAGED in CUmemAllocationType; older 12.x does not
    _allocation_type = {"pinned": _t.CU_MEM_ALLOCATION_TYPE_PINNED}  # noqa: RUF012
    if binding_version() >= (13, 0, 0):
        _allocation_type["managed"] = _t.CU_MEM_ALLOCATION_TYPE_MANAGED

    @staticmethod
    def _access_to_flags(spec: str):
        flags = VirtualMemoryResourceOptions._access_flags.get(spec)
        if flags is None:
            raise ValueError(f"Unknown access spec: {spec!r}")
        return flags

    @staticmethod
    def _allocation_type_to_driver(spec: str):
        alloc_type = VirtualMemoryResourceOptions._allocation_type.get(spec)
        if alloc_type is None:
            raise ValueError(f"Unsupported allocation_type: {spec!r}")
        return alloc_type

    @staticmethod
    def _location_type_to_driver(spec: str):
        loc_type = VirtualMemoryResourceOptions._location_type.get(spec)
        if loc_type is None:
            raise ValueError(f"Unsupported location_type: {spec!r}")
        return loc_type

    @staticmethod
    def _handle_type_to_driver(spec: str):
        if spec == "win32":
            raise NotImplementedError("win32 is currently not supported, please reach out to the CUDA Python team")
        handle_type = VirtualMemoryResourceOptions._handle_types.get(spec)
        if handle_type is None:
            raise ValueError(f"Unsupported handle_type: {spec!r}")
        return handle_type

    @staticmethod
    def _granularity_to_driver(spec: str):
        granularity = VirtualMemoryResourceOptions._granularity.get(spec)
        if granularity is None:
            raise ValueError(f"Unsupported granularity: {spec!r}")
        return granularity


class VirtualMemoryResource(MemoryResource):
    """Create a device memory resource that uses the CUDA VMM APIs to allocate memory.

    Parameters
    ----------
    device_id : Device | int
        Device for which a memory resource is constructed.

    config : VirtualMemoryResourceOptions
        A configuration object for the VirtualMemoryResource


    Warning
    -------
        This is a low-level API that is provided only for convenience. Make sure you fully understand
        how CUDA Virtual Memory Management works before using this. Other MemoryResource subclasses
        in cuda.core should already meet the common needs.
    """

    def __init__(self, device_id: Device | int, config: VirtualMemoryResourceOptions = None):
        self.device = Device(device_id)
        self.config = check_or_create_options(
            VirtualMemoryResourceOptions, config, "VirtualMemoryResource options", keep_none=False
        )
        # Matches ("host", "host_numa", "host_numa_current")
        if "host" in self.config.location_type:
            self.device = None

        if not self.device and self.config.location_type == "device":
            raise RuntimeError("VirtualMemoryResource requires a device for device memory allocations")

        if self.device and not self.device.properties.virtual_memory_management_supported:
            raise RuntimeError("VirtualMemoryResource requires CUDA VMM API support")

        # Validate RDMA support if requested
        if (
            self.config.gpu_direct_rdma
            and self.device is not None
            and not self.device.properties.gpu_direct_rdma_supported
        ):
            raise RuntimeError("GPU Direct RDMA is not supported on this device")

    @staticmethod
    def _align_up(size: int, gran: int) -> int:
        """
        Align a size up to the nearest multiple of a granularity.
        """
        return (size + gran - 1) & ~(gran - 1)

    def modify_allocation(self, buf: Buffer, new_size: int, config: VirtualMemoryResourceOptions = None) -> Buffer:
        """
        Grow an existing allocation using CUDA VMM, with a configurable policy.

        This implements true growing allocations that preserve the base pointer
        by extending the virtual address range and mapping additional physical memory.

        This function uses transactional allocation: if any step fails, the original buffer is not modified and
        all steps the function took are rolled back so a new allocation is not created.

        Parameters
        ----------
        buf : Buffer
            The existing buffer to grow
        new_size : int
            The new total size for the allocation
        config : VirtualMemoryResourceOptions, optional
            Configuration for the new physical memory chunks. If None, uses current config.

        Returns
        -------
        Buffer
            The same buffer with updated size and properties, preserving the original pointer
        """
        if config is not None:
            self.config = config

        # Build allocation properties for new chunks
        prop = driver.CUmemAllocationProp()
        prop.type = VirtualMemoryResourceOptions._allocation_type_to_driver(self.config.allocation_type)
        prop.location.type = VirtualMemoryResourceOptions._location_type_to_driver(self.config.location_type)
        prop.location.id = self.device.device_id
        prop.allocFlags.gpuDirectRDMACapable = 1 if self.config.gpu_direct_rdma else 0
        prop.requestedHandleTypes = VirtualMemoryResourceOptions._handle_type_to_driver(self.config.handle_type)
        prop.win32HandleMetaData = 0

        # Query granularity
        gran_flag = VirtualMemoryResourceOptions._granularity_to_driver(self.config.granularity)
        res, gran = driver.cuMemGetAllocationGranularity(prop, gran_flag)
        raise_if_driver_error(res)

        # Calculate sizes
        additional_size = new_size - buf.size
        if additional_size <= 0:
            # Same size: only update access policy if needed; avoid zero-sized driver calls
            descs = self._build_access_descriptors(prop)
            if descs:
                (res,) = driver.cuMemSetAccess(int(buf.handle), buf.size, descs, len(descs))
                raise_if_driver_error(res)
            return buf

        aligned_additional_size = VirtualMemoryResource._align_up(additional_size, gran)
        total_aligned_size = VirtualMemoryResource._align_up(new_size, gran)
        aligned_prev_size = total_aligned_size - aligned_additional_size
        addr_align = self.config.addr_align or gran

        # Try to extend the existing VA range first
        res, new_ptr = driver.cuMemAddressReserve(
            aligned_additional_size,
            addr_align,
            int(buf.handle) + aligned_prev_size,  # fixedAddr hint - aligned end of current range
            0,
        )

        if res != driver.CUresult.CUDA_SUCCESS or new_ptr != (int(buf.handle) + aligned_prev_size):
            # Check for specific errors that are not recoverable with the slow path
            if res in (
                driver.CUresult.CUDA_ERROR_INVALID_VALUE,
                driver.CUresult.CUDA_ERROR_NOT_PERMITTED,
                driver.CUresult.CUDA_ERROR_NOT_INITIALIZED,
                driver.CUresult.CUDA_ERROR_NOT_SUPPORTED,
            ):
                raise_if_driver_error(res)
            (res2,) = driver.cuMemAddressFree(new_ptr, aligned_additional_size)
            raise_if_driver_error(res2)
            # Fallback: couldn't extend contiguously, need full remapping
            return self._grow_allocation_slow_path(
                buf, new_size, prop, aligned_additional_size, total_aligned_size, addr_align
            )
        else:
            # Success! We can extend the VA range contiguously
            return self._grow_allocation_fast_path(buf, new_size, prop, aligned_additional_size, new_ptr)

    def _grow_allocation_fast_path(
        self, buf: Buffer, new_size: int, prop: driver.CUmemAllocationProp, aligned_additional_size: int, new_ptr: int
    ) -> Buffer:
        """
        Fast path for growing a virtual memory allocation when the new region can be
        reserved contiguously after the existing buffer.

        This function creates and maps new physical memory for the additional size,
        sets access permissions, and updates the buffer size in place (the pointer
        remains unchanged).

        Args:
            buf (Buffer):
                The buffer to grow.

            new_size (int):
                The new total size in bytes.

            prop (driver.CUmemAllocationProp):
                Allocation properties for the new memory.

            aligned_additional_size (int):
                The size of the new region to allocate, aligned to granularity.

            new_ptr (int):
                The address of the newly reserved contiguous VA region (should
                be at the end of the current buffer).

        Returns:
            Buffer: The same buffer object with its size updated to `new_size`.
        """
        with Transaction() as trans:
            # Create new physical memory for the additional size
            trans.append(
                lambda np=new_ptr, s=aligned_additional_size: raise_if_driver_error(driver.cuMemAddressFree(np, s)[0])
            )
            res, new_handle = driver.cuMemCreate(aligned_additional_size, prop, 0)
            raise_if_driver_error(res)
            # Register undo for creation
            trans.append(lambda h=new_handle: raise_if_driver_error(driver.cuMemRelease(h)[0]))

            # Map the new physical memory to the extended VA range
            (res,) = driver.cuMemMap(new_ptr, aligned_additional_size, 0, new_handle, 0)
            raise_if_driver_error(res)
            # Register undo for mapping
            trans.append(
                lambda np=new_ptr, s=aligned_additional_size: raise_if_driver_error(driver.cuMemUnmap(np, s)[0])
            )

            # Set access permissions for the new portion
            descs = self._build_access_descriptors(prop)
            if descs:
                (res,) = driver.cuMemSetAccess(new_ptr, aligned_additional_size, descs, len(descs))
                raise_if_driver_error(res)

            # All succeeded, cancel undo actions
            trans.commit()

        # Update the buffer size (pointer stays the same)
        buf._size = new_size
        return buf

    def _grow_allocation_slow_path(
        self,
        buf: Buffer,
        new_size: int,
        prop: driver.CUmemAllocationProp,
        aligned_additional_size: int,
        total_aligned_size: int,
        addr_align: int,
    ) -> Buffer:
        """
        Slow path for growing a virtual memory allocation when the new region cannot be
        reserved contiguously after the existing buffer.

        This function reserves a new, larger virtual address (VA) range, remaps the old
        physical memory to the beginning of the new VA range, creates and maps new physical
        memory for the additional size, sets access permissions, and updates the buffer's
        pointer and size.

        Args:
            buf (Buffer): The buffer to grow.
            new_size (int): The new total size in bytes.
            prop (driver.CUmemAllocationProp): Allocation properties for the new memory.
            aligned_additional_size (int): The size of the new region to allocate, aligned to granularity.
            total_aligned_size (int): The total new size to reserve, aligned to granularity.
            addr_align (int): The required address alignment for the new VA range.

        Returns:
            Buffer: The buffer object updated with the new pointer and size.
        """
        with Transaction() as trans:
            # Reserve a completely new, larger VA range
            res, new_ptr = driver.cuMemAddressReserve(total_aligned_size, addr_align, 0, 0)
            raise_if_driver_error(res)
            # Register undo for VA reservation
            trans.append(
                lambda np=new_ptr, s=total_aligned_size: raise_if_driver_error(driver.cuMemAddressFree(np, s)[0])
            )

            # Get the old allocation handle for remapping
            result, old_handle = driver.cuMemRetainAllocationHandle(buf.handle)
            raise_if_driver_error(result)
            # Register undo for old_handle
            trans.append(lambda h=old_handle: raise_if_driver_error(driver.cuMemRelease(h)[0]))

            # Unmap the old VA range (aligned previous size)
            aligned_prev_size = total_aligned_size - aligned_additional_size
            (result,) = driver.cuMemUnmap(int(buf.handle), aligned_prev_size)
            raise_if_driver_error(result)

            def _remap_old():
                # Try to remap the old physical memory back to the original VA range
                try:
                    (res,) = driver.cuMemMap(int(buf.handle), aligned_prev_size, 0, old_handle, 0)
                    raise_if_driver_error(res)
                except Exception:  # noqa: S110
                    # TODO: consider logging this exception
                    pass

            trans.append(_remap_old)

            # Remap the old physical memory to the new VA range (aligned previous size)
            (res,) = driver.cuMemMap(int(new_ptr), aligned_prev_size, 0, old_handle, 0)
            raise_if_driver_error(res)

            # Register undo for mapping
            trans.append(lambda np=new_ptr, s=aligned_prev_size: raise_if_driver_error(driver.cuMemUnmap(np, s)[0]))

            # Create new physical memory for the additional size
            res, new_handle = driver.cuMemCreate(aligned_additional_size, prop, 0)
            raise_if_driver_error(res)

            # Register undo for new physical memory
            trans.append(lambda h=new_handle: raise_if_driver_error(driver.cuMemRelease(h)[0]))

            # Map the new physical memory to the extended portion (aligned offset)
            (res,) = driver.cuMemMap(int(new_ptr) + aligned_prev_size, aligned_additional_size, 0, new_handle, 0)
            raise_if_driver_error(res)

            # Register undo for mapping
            trans.append(
                lambda base=int(new_ptr), offs=aligned_prev_size, s=aligned_additional_size: raise_if_driver_error(
                    driver.cuMemUnmap(base + offs, s)[0]
                )
            )

            # Set access permissions for the entire new range
            descs = self._build_access_descriptors(prop)
            if descs:
                (res,) = driver.cuMemSetAccess(new_ptr, total_aligned_size, descs, len(descs))
                raise_if_driver_error(res)

            # All succeeded, cancel undo actions
            trans.commit()

        # Free the old VA range (aligned previous size)
        (res2,) = driver.cuMemAddressFree(int(buf.handle), aligned_prev_size)
        raise_if_driver_error(res2)

        # Invalidate the old buffer so its destructor won't try to free again
        buf._clear()

        # Return a new Buffer for the new mapping
        return Buffer.from_handle(ptr=new_ptr, size=new_size, mr=self)

    def _build_access_descriptors(self, prop: driver.CUmemAllocationProp) -> list:
        """
        Build access descriptors for memory access permissions.

        Returns
        -------
        list
            List of CUmemAccessDesc objects for setting memory access
        """
        descs = []

        # Owner access
        owner_flags = VirtualMemoryResourceOptions._access_to_flags(self.config.self_access)
        if owner_flags:
            d = driver.CUmemAccessDesc()
            d.location.type = prop.location.type
            d.location.id = prop.location.id
            d.flags = owner_flags
            descs.append(d)

        # Peer device access
        peer_flags = VirtualMemoryResourceOptions._access_to_flags(self.config.peer_access)
        if peer_flags:
            for peer_dev in self.config.peers:
                d = driver.CUmemAccessDesc()
                d.location.type = driver.CUmemLocationType.CU_MEM_LOCATION_TYPE_DEVICE
                d.location.id = int(peer_dev)
                d.flags = peer_flags
                descs.append(d)

        return descs

    def allocate(self, size: int, stream: Stream | None = None) -> Buffer:
        """
        Allocate a buffer of the given size using CUDA virtual memory.

        Parameters
        ----------
        size : int
            The size in bytes of the buffer to allocate.
        stream : Stream, optional
            CUDA stream to associate with the allocation (not currently supported).

        Returns
        -------
        Buffer
            A Buffer object representing the allocated virtual memory.

        Raises
        ------
        NotImplementedError
            If a stream is provided or if the location type is not device memory.
        CUDAError
            If any CUDA driver API call fails during allocation.

        Notes
        -----
        This method uses transactional allocation: if any step fails, all resources
        allocated so far are automatically cleaned up. The allocation is performed
        with the configured granularity, access permissions, and peer access as
        specified in the resource's configuration.
        """
        if stream is not None:
            raise NotImplementedError("Stream is not supported with VirtualMemoryResource")

        config = self.config
        # ---- Build allocation properties ----
        prop = driver.CUmemAllocationProp()
        prop.type = VirtualMemoryResourceOptions._allocation_type_to_driver(config.allocation_type)
        prop.location.type = VirtualMemoryResourceOptions._location_type_to_driver(config.location_type)
        prop.location.id = self.device.device_id if config.location_type == "device" else -1
        prop.allocFlags.gpuDirectRDMACapable = 1 if config.gpu_direct_rdma else 0
        prop.requestedHandleTypes = VirtualMemoryResourceOptions._handle_type_to_driver(config.handle_type)
        prop.win32HandleMetaData = 0

        # ---- Query and apply granularity ----
        # Choose min vs recommended granularity per config
        gran_flag = VirtualMemoryResourceOptions._granularity_to_driver(config.granularity)
        res, gran = driver.cuMemGetAllocationGranularity(prop, gran_flag)
        raise_if_driver_error(res)

        aligned_size = VirtualMemoryResource._align_up(size, gran)
        addr_align = config.addr_align or gran

        # ---- Transactional allocation ----
        with Transaction() as trans:
            # ---- Create physical memory ----
            res, handle = driver.cuMemCreate(aligned_size, prop, 0)
            raise_if_driver_error(res)
            # Register undo for physical memory
            trans.append(lambda h=handle: raise_if_driver_error(driver.cuMemRelease(h)[0]))

            # ---- Reserve VA space ----
            # Potentially, use a separate size for the VA reservation from the physical allocation size
            res, ptr = driver.cuMemAddressReserve(aligned_size, addr_align, config.addr_hint, 0)
            raise_if_driver_error(res)
            # Register undo for VA reservation
            trans.append(lambda p=ptr, s=aligned_size: raise_if_driver_error(driver.cuMemAddressFree(p, s)[0]))

            # ---- Map physical memory into VA ----
            (res,) = driver.cuMemMap(ptr, aligned_size, 0, handle, 0)
            trans.append(lambda p=ptr, s=aligned_size: raise_if_driver_error(driver.cuMemUnmap(p, s)[0]))
            raise_if_driver_error(res)

            # ---- Set access for owner + peers ----
            descs = self._build_access_descriptors(prop)
            if descs:
                (res,) = driver.cuMemSetAccess(ptr, aligned_size, descs, len(descs))
                raise_if_driver_error(res)

            trans.commit()

        # Done — return a Buffer that tracks this VA range
        buf = Buffer.from_handle(ptr=ptr, size=aligned_size, mr=self)
        return buf

    def deallocate(self, ptr: int, size: int, stream: Stream | None = None) -> None:  # noqa: ARG002
        """
        Deallocate memory on the device using CUDA VMM APIs.
        """
        result, handle = driver.cuMemRetainAllocationHandle(ptr)
        raise_if_driver_error(result)
        (result,) = driver.cuMemUnmap(ptr, size)
        raise_if_driver_error(result)
        (result,) = driver.cuMemAddressFree(ptr, size)
        raise_if_driver_error(result)
        (result,) = driver.cuMemRelease(handle)
        raise_if_driver_error(result)

    @property
    def is_device_accessible(self) -> bool:
        """
        Indicates whether the allocated memory is accessible from the device.
        """
        return self.config.location_type == "device"

    @property
    def is_host_accessible(self) -> bool:
        """
        Indicates whether the allocated memory is accessible from the host.
        """
        return self.config.location_type == "host"

    @property
    def device_id(self) -> int:
        """
        Get the device ID associated with this memory resource.

        Returns:
            int: CUDA device ID. -1 if the memory resource allocates host memory
        """
        return self.device.device_id if self.config.location_type == "device" else -1

    def __repr__(self) -> str:
        """
        Return a string representation of the VirtualMemoryResource.

        Returns:
            str: A string describing the object
        """
        return f"<VirtualMemoryResource device={self.device}>"



# SPDX-FileCopyrightText: Copyright (c) 2024-2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
#
# SPDX-License-Identifier: Apache-2.0

from cuda.core._resource_handles cimport ContextHandle

cdef class Context:
    """Cython declaration for Context class.

    This class provides access to CUDA contexts. Context objects cannot be
    instantiated directly - use factory methods or Device/Stream APIs.
    """

    cdef:
        ContextHandle _h_context
        int _device_id
        object __weakref__

    @staticmethod
    cdef Context _from_handle(type cls, ContextHandle h_context, int device_id)



# SPDX-FileCopyrightText: Copyright (c) 2024-2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
#
# SPDX-License-Identifier: Apache-2.0

from ._linker cimport Linker
from ._resource_handles cimport NvrtcProgramHandle, NvvmProgramHandle


cdef class Program:
    cdef:
        NvrtcProgramHandle _h_nvrtc
        NvvmProgramHandle _h_nvvm
        str _backend
        Linker _linker
        object _options  # ProgramOptions
        object __weakref__
        object _compile_lock  # Per-instance lock for compile-time mutation
        bint _use_libdevice      # Flag for libdevice loading
        bint _libdevice_added
        bytes _nvrtc_code       # Source code for NVRTC retry (PCH auto-resize)
        str _pch_status         # PCH creation outcome after compile



# SPDX-FileCopyrightText: Copyright (c) 2024-2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
#
# SPDX-License-Identifier: Apache-2.0

from cuda.core._version import __version__

try:
    from cuda import bindings
except ImportError:
    raise ImportError("cuda.bindings 12.x or 13.x must be installed") from None
else:
    cuda_major, cuda_minor = bindings.__version__.split(".")[:2]
    if cuda_major not in ("12", "13"):
        raise ImportError("cuda.bindings 12.x or 13.x must be installed")

import importlib

subdir = f"cu{cuda_major}"
try:
    versioned_mod = importlib.import_module(f".{subdir}", __package__)
    # Import all symbols from the module
    globals().update(versioned_mod.__dict__)
except ImportError:
    # This is not a wheel build, but a conda or local build, do nothing
    pass
else:
    del versioned_mod
finally:
    del bindings, importlib, subdir, cuda_major, cuda_minor

from cuda.core import system, utils
from cuda.core._device import Device
from cuda.core._event import Event, EventOptions
from cuda.core._graphics import GraphicsResource
from cuda.core._launch_config import LaunchConfig
from cuda.core._launcher import launch
from cuda.core._layout import _StridedLayout
from cuda.core._linker import Linker, LinkerOptions
from cuda.core._memory import (
    Buffer,
    DeviceMemoryResource,
    DeviceMemoryResourceOptions,
    GraphMemoryResource,
    LegacyPinnedMemoryResource,
    ManagedMemoryResource,
    ManagedMemoryResourceOptions,
    MemoryResource,
    PinnedMemoryResource,
    PinnedMemoryResourceOptions,
    VirtualMemoryResource,
    VirtualMemoryResourceOptions,
)
from cuda.core._memoryview import (
    StridedMemoryView,
    args_viewable_as_strided_memory,
)
from cuda.core._module import Kernel, ObjectCode
from cuda.core._program import Program, ProgramOptions
from cuda.core._stream import (
    LEGACY_DEFAULT_STREAM,
    PER_THREAD_DEFAULT_STREAM,
    Stream,
    StreamOptions,
)
from cuda.core._tensor_map import TensorMapDescriptor, TensorMapDescriptorOptions
from cuda.core.graph import (
    Condition,
    Graph,
    GraphAllocOptions,
    GraphBuilder,
    GraphCompleteOptions,
    GraphDebugPrintOptions,
    GraphDef,
)



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
#
# SPDX-License-Identifier: Apache-2.0

from libcpp.vector cimport vector

from cuda.bindings cimport cydriver


cdef class LaunchConfig:
    """Customizable launch options."""
    cdef:
        public tuple grid
        public tuple cluster
        public tuple block
        public int shmem_size
        public bint cooperative_launch

        vector[cydriver.CUlaunchAttribute] _attrs
        object __weakref__

    cdef cydriver.CUlaunchConfig _to_native_launch_config(self)


cpdef object _to_native_launch_config(LaunchConfig config)



# SPDX-FileCopyrightText: Copyright (c) 2024-2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
#
# SPDX-License-Identifier: Apache-2.0

from cuda.bindings cimport cydriver
from cuda.core._resource_handles cimport LibraryHandle, KernelHandle

cdef class ObjectCode
cdef class Kernel
cdef class KernelOccupancy
cdef class KernelAttributes


cdef class Kernel:
    cdef:
        KernelHandle _h_kernel
        KernelAttributes _attributes  # lazy
        KernelOccupancy _occupancy  # lazy
        object _keepalive
        object __weakref__

    @staticmethod
    cdef Kernel _from_handle(KernelHandle h_kernel)

    cdef tuple _get_arguments_info(self, bint param_info=*)


cdef class ObjectCode:
    cdef:
        LibraryHandle _h_library
        str _code_type
        object _module      # bytes/str source
        dict _sym_map
        str _name
        object __weakref__

    cdef int _lazy_load_module(self) except -1


cdef class KernelOccupancy:
    cdef:
        KernelHandle _h_kernel

    @staticmethod
    cdef KernelOccupancy _init(KernelHandle h_kernel)


cdef class KernelAttributes:
    cdef:
        KernelHandle _h_kernel
        dict _cache

    @staticmethod
    cdef KernelAttributes _init(KernelHandle h_kernel)

    cdef int _get_cached_attribute(self, int device_id, cydriver.CUfunction_attribute attribute) except? -1
    cdef int _resolve_device_id(self, device_id) except? -1



# SPDX-FileCopyrightText: Copyright (c) 2024-2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
#
# SPDX-License-Identifier: Apache-2.0

from cuda.bindings cimport cydriver
from cuda.core._memory._buffer cimport Buffer
from cuda.core._memory._memory_pool cimport _MemPool
from cuda.core._resource_handles cimport FileDescriptorHandle


# Holds _MemPool objects imported by this process.  This enables
# buffer serialization, as buffers can reduce to a pair comprising the memory
# resource UUID (the key into this registry) and the serialized buffer
# descriptor.
cdef object registry


# The IPC handle type for this platform.  IPC is currently only supported on
# Linux. On other platforms, the IPC handle type is set equal to the no-IPC
# handle type.
cdef cydriver.CUmemAllocationHandleType IPC_HANDLE_TYPE


# Whether IPC is supported on this platform.
cdef is_supported()


cdef class IPCDataForBuffer:
    cdef:
        IPCBufferDescriptor _ipc_descriptor
        bint                _is_mapped


cdef class IPCDataForMR:
    cdef:
        IPCAllocationHandle _alloc_handle
        bint                _is_mapped


cdef class IPCBufferDescriptor:
    cdef:
        bytes  _payload
        size_t _size

    cdef const void* payload_ptr(self) noexcept


cdef class IPCAllocationHandle:
    cdef:
        FileDescriptorHandle _h_fd
        object               _uuid

    cpdef close(self)


# Buffer IPC Implementation
# -------------------------
cdef IPCBufferDescriptor Buffer_get_ipc_descriptor(Buffer)
cdef Buffer Buffer_from_ipc_descriptor(cls, _MemPool, IPCBufferDescriptor, stream)


# _MemPool IPC Implementation
# ---------------------------
cdef _MemPool MP_from_allocation_handle(cls, alloc_handle)
cdef _MemPool MP_from_registry(uuid)
cdef _MemPool MP_register(_MemPool, uuid)
cdef IPCAllocationHandle MP_export_mempool(_MemPool)



# SPDX-FileCopyrightText: Copyright (c) 2025-2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
#
# SPDX-License-Identifier: Apache-2.0

from cuda.bindings cimport cydriver
from cuda.core._resource_handles cimport GraphHandle, GraphNodeHandle


cdef class GraphNode:
    cdef:
        GraphNodeHandle _h_node
        object __weakref__

    @staticmethod
    cdef GraphNode _create(GraphHandle h_graph, cydriver.CUgraphNode node)



# SPDX-FileCopyrightText: Copyright (c) 2025-2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
#
# SPDX-License-Identifier: Apache-2.0


from cuda.bindings import nvml

NvmlError = nvml.NvmlError
UninitializedError = nvml.UninitializedError
InvalidArgumentError = nvml.InvalidArgumentError
NotSupportedError = nvml.NotSupportedError
NoPermissionError = nvml.NoPermissionError
AlreadyInitializedError = nvml.AlreadyInitializedError
NotFoundError = nvml.NotFoundError
InsufficientSizeError = nvml.InsufficientSizeError
InsufficientPowerError = nvml.InsufficientPowerError
DriverNotLoadedError = nvml.DriverNotLoadedError
TimeoutError = nvml.TimeoutError
IrqIssueError = nvml.IrqIssueError
LibraryNotFoundError = nvml.LibraryNotFoundError
FunctionNotFoundError = nvml.FunctionNotFoundError
CorruptedInforomError = nvml.CorruptedInforomError
GpuIsLostError = nvml.GpuIsLostError
ResetRequiredError = nvml.ResetRequiredError
OperatingSystemError = nvml.OperatingSystemError
LibRmVersionMismatchError = nvml.LibRmVersionMismatchError
InUseError = nvml.InUseError
MemoryError = nvml.MemoryError
NoDataError = nvml.NoDataError
VgpuEccNotSupportedError = nvml.VgpuEccNotSupportedError
InsufficientResourcesError = nvml.InsufficientResourcesError
FreqNotSupportedError = nvml.FreqNotSupportedError
ArgumentVersionMismatchError = nvml.ArgumentVersionMismatchError
DeprecatedError = nvml.DeprecatedError
NotReadyError = nvml.NotReadyError
GpuNotFoundError = nvml.GpuNotFoundError
InvalidStateError = nvml.InvalidStateError
ResetTypeNotSupportedError = nvml.ResetTypeNotSupportedError
UnknownError = nvml.UnknownError


__all__ = [
    "AlreadyInitializedError",
    "ArgumentVersionMismatchError",
    "CorruptedInforomError",
    "DeprecatedError",
    "DriverNotLoadedError",
    "FreqNotSupportedError",
    "FunctionNotFoundError",
    "GpuIsLostError",
    "GpuNotFoundError",
    "InUseError",
    "InsufficientPowerError",
    "InsufficientResourcesError",
    "InsufficientSizeError",
    "InvalidArgumentError",
    "InvalidStateError",
    "IrqIssueError",
    "LibRmVersionMismatchError",
    "LibraryNotFoundError",
    "MemoryError",
    "NoDataError",
    "NoPermissionError",
    "NotFoundError",
    "NotReadyError",
    "NotSupportedError",
    "NvmlError",
    "OperatingSystemError",
    "ResetRequiredError",
    "ResetTypeNotSupportedError",
    "TimeoutError",
    "UninitializedError",
    "UnknownError",
    "VgpuEccNotSupportedError",
]



# SPDX-FileCopyrightText: Copyright (c) 2024-2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
#
# SPDX-License-Identifier: Apache-2.0

from libc.stdint cimport intptr_t

from cuda.core._dlpack cimport DLTensor
from cuda.core._layout cimport _StridedLayout


cdef class StridedMemoryView:
    cdef readonly:
        intptr_t ptr
        int device_id
        bint is_device_accessible
        bint readonly
        object exporting_obj

    cdef:
        object metadata
        DLTensor* dl_tensor
        _StridedLayout _layout
        object _buffer
        object _dtype

    cdef inline _StridedLayout get_layout(self)
    cdef inline object get_buffer(self)
    cdef inline object get_dtype(self)



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
#
# SPDX-License-Identifier: Apache-2.0


cdef extern from *:
    """
    #if defined(_WIN32) || defined(_WIN64)
        #include <process.h>
    #else
        #include <unistd.h>
    #endif
    """
    int getpid() nogil


ctypedef enum _NVMLState:
    UNINITIALIZED = 0
    INITIALIZED = 1
    DISABLED_LIBRARY_NOT_FOUND = 2


# Initialisation must occur per-process, so an initialised state is a
# (state, pid) pair
cdef _NVMLState _NVML_STATE


cdef int _NVML_OWNER_PID


cpdef _initialize()


cpdef inline initialize():
    """
    Initializes Nvidia Management Library (NVML), ensuring it only happens once per process.
    """
    if _NVML_STATE == _NVMLState.DISABLED_LIBRARY_NOT_FOUND or (
        _NVML_STATE == _NVMLState.INITIALIZED and getpid() == _NVML_OWNER_PID
    ):
        return

    _initialize()


cpdef inline bint is_initialized():
    """
    Check whether the NVML context is initialized on this process.

    Returns
    -------
    result: bool
        Whether the NVML context is initialized on this process.
    """
    return _NVML_STATE == _NVMLState.INITIALIZED and getpid() == _NVML_OWNER_PID


cpdef validate()



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
#
# SPDX-License-Identifier: Apache-2.0


# NOTE: We must maintain that it is always possible to import this module
# without CUDA being installed, and without CUDA being initialized or any
# contexts created, so that a user can use NVML to explore things about their
# system without loading CUDA.


__all__ = [
    "CUDA_BINDINGS_NVML_IS_COMPATIBLE",
    "get_driver_version",
    "get_driver_version_full",
    "get_num_devices",
    "get_process_name",
]


from ._system import *

if CUDA_BINDINGS_NVML_IS_COMPATIBLE:
    from ._device import *
    from ._device import __all__ as _device_all
    from ._system_events import *
    from ._system_events import __all__ as _system_events_all
    from .exceptions import *
    from .exceptions import __all__ as _exceptions_all

    __all__.append("get_nvml_version")
    __all__.extend(_device_all)
    __all__.extend(_system_events_all)
    __all__.extend(_exceptions_all)



# SPDX-FileCopyrightText: Copyright (c) 2024-2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
#
# SPDX-License-Identifier: Apache-2.0

from ._resource_handles cimport NvJitLinkHandle, CuLinkHandle


cdef class Linker:
    cdef:
        NvJitLinkHandle _nvjitlink_handle
        CuLinkHandle _culink_handle
        bint _use_nvjitlink
        object _drv_log_bufs  # formatted_options list (driver); None for nvjitlink; cleared in link()
        str _info_log         # decoded log; None until link() or pre-link get_*_log()
        str _error_log        # decoded log; None until link() or pre-link get_*_log()
        object _options       # LinkerOptions
        object __weakref__



# SPDX-FileCopyrightText: Copyright (c) 2025-2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
#
# SPDX-License-Identifier: Apache-2.0

from cuda.bindings cimport cydriver
from cuda.core._resource_handles cimport GraphHandle


cdef class Condition:
    cdef:
        cydriver.CUgraphConditionalHandle _c_handle
        object __weakref__


cdef class GraphDef:
    cdef:
        GraphHandle _h_graph
        object __weakref__

    @staticmethod
    cdef GraphDef _from_handle(GraphHandle h_graph)



# SPDX-FileCopyrightText: Copyright (c) 2024-2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
#
# SPDX-License-Identifier: Apache-2.0

from libc.stdint cimport intptr_t
from libcpp cimport vector


cdef class ParamHolder:

    cdef:
        vector.vector[void*] data
        vector.vector[void*] data_addresses
        object kernel_args
        readonly intptr_t ptr



# SPDX-FileCopyrightText: Copyright (c) 2025-2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
#
# SPDX-License-Identifier: Apache-2.0

from cuda.core._resource_handles cimport GraphicsResourceHandle


cdef class GraphicsResource:

    cdef:
        GraphicsResourceHandle _handle
        object _mapped_buffer
        object _context_manager_stream
        object _entered_buffer

    cpdef close(self, stream=*)



# SPDX-FileCopyrightText: Copyright (c) 2024-2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
#
# SPDX-License-Identifier: Apache-2.0

from cuda.core._memory._memory_pool cimport _MemPool


cdef class ManagedMemoryResource(_MemPool):
    cdef:
        str _pref_loc_type
        int _pref_loc_id



# SPDX-FileCopyrightText: Copyright (c) 2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
#
# SPDX-License-Identifier: Apache-2.0

cdef tuple cy_binding_version()
cdef tuple cy_driver_version()



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
#
# SPDX-License-Identifier: Apache-2.0

cimport cython
from cython.operator cimport dereference as deref

from libc.stdint cimport int64_t, int32_t, uint32_t, uint64_t, intptr_t
from libcpp cimport vector

ctypedef int64_t extent_t
ctypedef int64_t stride_t
ctypedef int32_t axis_t

ctypedef uint64_t axes_mask_t  # MUST be exactly STRIDED_LAYOUT_MAX_NDIM bits wide
ctypedef uint32_t property_mask_t

ctypedef vector.vector[stride_t] extents_strides_t
ctypedef vector.vector[axis_t] axis_vec_t

from cuda.core._utils cimport cuda_utils


ctypedef fused integer_t:
    int64_t
    int32_t


cdef extern from "_include/layout.hpp":

    cdef int STRIDED_LAYOUT_MAX_NDIM
    cdef axes_mask_t AXES_MASK_ALL
    int64_t _c_abs(int64_t x) nogil
    void _order_from_strides(axis_vec_t& indices, extent_t* extent_t, stride_t* stride_t, int ndim) except + nogil
    void _swap(extents_strides_t &a, extents_strides_t &b) noexcept nogil
    void _swap(int64_t* a, int64_t* b) noexcept nogil
    void _swap(int a, int b) noexcept nogil
    void _swap(axis_vec_t &a, axis_vec_t &b) noexcept nogil


cdef enum OrderFlag:
    ORDER_NONE = 0
    ORDER_C = 1
    ORDER_F = 2
    ORDER_PERM = 3


cdef enum Property:
    PROP_IS_UNIQUE = 1 << 0
    PROP_IS_CONTIGUOUS_C = 1 << 1
    PROP_IS_CONTIGUOUS_F = 1 << 2
    PROP_IS_CONTIGUOUS_ANY = 1 << 3
    PROP_IS_DENSE = 1 << 4
    PROP_OFFSET_BOUNDS = 1 << 5
    PROP_REQUIRED_SIZE_IN_BYTES = 1 << 6
    PROP_SHAPE = 1 << 7
    PROP_STRIDES = 1 << 8
    PROP_STRIDES_IN_BYTES = 1 << 9
    PROP_STRIDE_ORDER = 1 << 10
    PROP_VOLUME = 1 << 11


cdef struct BaseLayout:
    # A struct holding the shape and strides for the layout.
    # Use ``init_base_layout`` to initialize the layout, it will
    # set the ``shape`` and ``strides`` pointers to point to
    # ndim contigious integer arrays.
    # The ``shape`` pointer must not be NULL, the ``strides`` can be
    # set to NULL by the user to indicate C-contiguous layout.
    # Uses single _mem allocation to reduce overhead
    # (allocation and exceptions checks).

    extents_strides_t _mem
    extent_t* shape
    stride_t* strides
    int ndim


@cython.final
cdef class _StridedLayout:

    # Definition
    cdef:
        BaseLayout base

        readonly:
            int itemsize
            stride_t slice_offset

    # Lazy properties computed from the defining values.
    cdef:
        # Set to 0 to invalidate all properties,
        # whenever a defining value is changed
        property_mask_t _prop_mask

        # C and Python properties
        property_mask_t _boolean_props
        int64_t _required_size_in_bytes
        stride_t _min_offset
        stride_t _max_offset
        int64_t _volume

        # Python properties
        tuple _py_shape
        tuple _py_strides
        tuple _py_strides_in_bytes
        tuple _py_stride_order

    # ==============================
    # Initialization
    # ==============================

    cdef inline int _init(_StridedLayout self, BaseLayout& base, int itemsize, bint divide_strides=False) except -1 nogil:
        if itemsize <= 0:
            raise ValueError("itemsize must be positive")

        if base.strides != NULL and divide_strides:
            _divide_strides(base, itemsize)

        self.itemsize = itemsize
        self.slice_offset = 0

        _swap_layout(self.base, base)
        return 0

    cdef inline stride_t _init_dense(_StridedLayout self, BaseLayout& base, int itemsize, OrderFlag order_flag, axis_vec_t* stride_order=NULL) except -1 nogil:
        if itemsize <= 0:
            raise ValueError("itemsize must be positive")

        cdef stride_t volume
        if order_flag == ORDER_C:
            volume = _dense_strides_c(base)
        elif order_flag == ORDER_F:
            volume = _dense_strides_f(base)
        elif order_flag == ORDER_PERM:
            if stride_order == NULL:
                raise ValueError("stride_order is required for ORDER_PERM")
            volume = _dense_strides_in_order(base, deref(stride_order))
        else:
            raise ValueError("The stride_order must be 'C', 'F', or a permutation.")

        self.itemsize = itemsize
        self.slice_offset = 0
        _swap_layout(self.base, base)
        self._volume = volume
        _mark_property_valid(self, PROP_VOLUME)
        return 0

    cdef inline int init_from_ptr(_StridedLayout self, int ndim, extent_t* shape, stride_t* strides, int itemsize, bint divide_strides=False) except -1 nogil:
        cdef BaseLayout base
        _init_base_layout_from_ptr(base, ndim, shape, strides)
        return self._init(base, itemsize, divide_strides)

    cdef inline int init_dense_from_ptr(_StridedLayout self, int ndim, extent_t* shape, int itemsize, OrderFlag order_flag, axis_vec_t* stride_order=NULL) except -1 nogil:
        cdef BaseLayout base
        _init_base_layout_from_ptr(base, ndim, shape, NULL)
        return self._init_dense(base, itemsize, order_flag, stride_order)

    cdef inline int init_from_tuple(_StridedLayout self, tuple shape, tuple strides, int itemsize, bint divide_strides=False) except -1:
        cdef BaseLayout base
        _init_base_layout_from_tuple(base, shape, strides)
        return self._init(base, itemsize, divide_strides)

    cdef inline int init_dense_from_tuple(_StridedLayout self, tuple shape, int itemsize, object stride_order) except -1:
        cdef axis_vec_t stride_order_vec
        cdef OrderFlag order_flag = _stride_order2vec(stride_order_vec, stride_order)

        if order_flag == ORDER_NONE:
            raise ValueError(f"The stride_order must be 'C', 'F', or a permutation tuple. Got: {stride_order}")

        cdef BaseLayout base
        _init_base_layout_from_tuple(base, shape, None)
        return self._init_dense(base, itemsize, order_flag, &stride_order_vec)

    # ==============================
    # Properties
    # ==============================

    cdef inline tuple get_shape_tuple(_StridedLayout self):
        if not _has_valid_property(self, PROP_SHAPE):
            self._py_shape = cuda_utils.carray_integer_t_to_tuple(self.base.shape, self.base.ndim)
            _mark_property_valid(self, PROP_SHAPE)
        return self._py_shape

    cdef inline tuple get_strides_tuple(_StridedLayout self):
        if not _has_valid_property(self, PROP_STRIDES):
            if self.base.strides == NULL:
                self._py_strides = None
            else:
                self._py_strides = cuda_utils.carray_integer_t_to_tuple(self.base.strides, self.base.ndim)
            _mark_property_valid(self, PROP_STRIDES)
        return self._py_strides

    cdef inline int get_strides_in_bytes(_StridedLayout self, extents_strides_t& strides) except -1 nogil:
        if self.base.strides != NULL:
            strides.resize(self.base.ndim)
            for i in range(self.base.ndim):
                strides[i] = _overflow_checked_mul(self.base.strides[i], self.itemsize)
        return 0

    cdef inline tuple get_strides_in_bytes_tuple(_StridedLayout self):
        if _has_valid_property(self, PROP_STRIDES_IN_BYTES):
            return self._py_strides_in_bytes
        cdef extents_strides_t strides
        if self.base.strides == NULL:
            self._py_strides_in_bytes = None
        else:
            self.get_strides_in_bytes(strides)
            self._py_strides_in_bytes = cuda_utils.carray_integer_t_to_tuple(strides.data(), strides.size())
        _mark_property_valid(self, PROP_STRIDES_IN_BYTES)
        return self._py_strides_in_bytes

    cdef inline int64_t get_volume(_StridedLayout self) except -1 nogil:
        if not _has_valid_property(self, PROP_VOLUME):
            self._volume = _volume(self.base)
            _mark_property_valid(self, PROP_VOLUME)
        return self._volume

    cdef inline int get_stride_order(_StridedLayout self, axis_vec_t& stride_order) except -1 nogil:
        _order_from_strides(stride_order, self.base.shape, self.base.strides, self.base.ndim)
        return 0

    cdef inline tuple get_stride_order_tuple(_StridedLayout self):
        if _has_valid_property(self, PROP_STRIDE_ORDER):
            return self._py_stride_order
        cdef axis_vec_t stride_order
        self.get_stride_order(stride_order)
        self._py_stride_order = cuda_utils.carray_integer_t_to_tuple(stride_order.data(), stride_order.size())
        _mark_property_valid(self, PROP_STRIDE_ORDER)
        return self._py_stride_order

    cdef inline bint get_is_unique(_StridedLayout self) except -1 nogil:
        if _has_valid_property(self, PROP_IS_UNIQUE):
            return _boolean_property(self, PROP_IS_UNIQUE)
        if self.base.strides == NULL or self.get_volume() == 0:
            return _set_boolean_property(self, PROP_IS_UNIQUE, True)
        cdef axis_vec_t stride_order
        self.get_stride_order(stride_order)
        return _set_boolean_property(self, PROP_IS_UNIQUE, _is_unique(self.base, stride_order))

    cdef inline bint get_is_contiguous_c(_StridedLayout self) except -1 nogil:
        if _has_valid_property(self, PROP_IS_CONTIGUOUS_C):
            return _boolean_property(self, PROP_IS_CONTIGUOUS_C)
        return _set_boolean_property(self, PROP_IS_CONTIGUOUS_C, _is_contiguous_c(self.get_volume(), self.base))

    cdef inline bint get_is_contiguous_f(_StridedLayout self) except -1 nogil:
        if _has_valid_property(self, PROP_IS_CONTIGUOUS_F):
            return _boolean_property(self, PROP_IS_CONTIGUOUS_F)
        return _set_boolean_property(self, PROP_IS_CONTIGUOUS_F, _is_contiguous_f(self.get_volume(), self.base))

    cdef inline bint get_is_contiguous_any(_StridedLayout self) except -1 nogil:
        if _has_valid_property(self, PROP_IS_CONTIGUOUS_ANY):
            return _boolean_property(self, PROP_IS_CONTIGUOUS_ANY)
        cdef axis_vec_t stride_order
        self.get_stride_order(stride_order)
        return _set_boolean_property(self, PROP_IS_CONTIGUOUS_ANY, _is_contiguous_any(self.get_volume(), self.base, stride_order))

    cdef inline bint get_is_dense(_StridedLayout self) except -1 nogil:
        if _has_valid_property(self, PROP_IS_DENSE):
            return _boolean_property(self, PROP_IS_DENSE)
        return _set_boolean_property(self, PROP_IS_DENSE, self.slice_offset == 0 and self.get_is_contiguous_any())

    cdef inline int get_offset_bounds(_StridedLayout self, stride_t& min_offset, stride_t& max_offset) except -1 nogil:
        if _has_valid_property(self, PROP_OFFSET_BOUNDS):
            min_offset = self._min_offset
            max_offset = self._max_offset
            return 0
        cdef int ndim = self.base.ndim
        cdef stride_t stride
        cdef extent_t extent
        min_offset = self.slice_offset
        max_offset = self.slice_offset
        if self.base.strides == NULL:
            max_offset = _overflow_checked_sum(max_offset, self.get_volume() - 1)
        else:
            for i in range(ndim):
                stride = self.base.strides[i]  # can be negative
                extent = self.base.shape[i]  # must be non-negative
                if extent == 0:
                    min_offset = 0
                    max_offset = -1  # empty range
                    return 0
                if stride <= 0:
                    min_offset = _overflow_checked_sum(min_offset, _overflow_checked_mul(stride, (extent - 1)))
                else:
                    max_offset = _overflow_checked_sum(max_offset, _overflow_checked_mul(stride, (extent - 1)))
        self._min_offset = min_offset
        self._max_offset = max_offset
        _mark_property_valid(self, PROP_OFFSET_BOUNDS)
        return 0

    cdef inline int64_t get_required_size_in_bytes(_StridedLayout self) except? -1 nogil:
        if _has_valid_property(self, PROP_REQUIRED_SIZE_IN_BYTES):
            return self._required_size_in_bytes
        cdef stride_t min_offset = 0
        cdef stride_t max_offset = 0
        self.get_offset_bounds(min_offset, max_offset)
        if min_offset < 0:
            raise ValueError(
                f"Allocation size for a layout that maps elements "
                f"to negative memory offsets is ambiguous. "
                f"The layout's min_offset is {min_offset}."
            )
        if max_offset < min_offset:
            return 0
        cdef int64_t required_size_in_bytes = _overflow_checked_sum(max_offset, 1)
        self._required_size_in_bytes = _overflow_checked_mul(required_size_in_bytes, self.itemsize)
        _mark_property_valid(self, PROP_REQUIRED_SIZE_IN_BYTES)
        return self._required_size_in_bytes

    cdef inline int64_t get_slice_offset_in_bytes(_StridedLayout self) except? -1 nogil:
        return _overflow_checked_mul(self.slice_offset, self.itemsize)

    cdef axes_mask_t get_flattened_axis_mask(_StridedLayout self) except? -1 nogil
    cdef int get_max_compatible_itemsize(_StridedLayout self, int max_itemsize, intptr_t data_ptr, int axis=*) except -1 nogil

    # ==============================
    # Layout manipulation
    # ==============================


    cdef int reshape_into(_StridedLayout self, _StridedLayout out_layout, BaseLayout& new_shape) except -1 nogil
    cdef int permute_into(_StridedLayout self, _StridedLayout out_layout, axis_vec_t& axis_order) except -1 nogil

    cdef int flatten_into(_StridedLayout self, _StridedLayout out_layout, axes_mask_t axis_mask) except -1 nogil
    cdef int squeeze_into(_StridedLayout self, _StridedLayout out_layout) except -1 nogil
    cdef int unsqueeze_into(_StridedLayout self, _StridedLayout out_layout, axis_vec_t& axis_vec) except -1 nogil
    cdef int broadcast_into(_StridedLayout self, _StridedLayout out_layout, BaseLayout& broadcast) except -1 nogil
    cdef int pack_into(_StridedLayout self, _StridedLayout out_layout, int itemsize, intptr_t data_ptr, bint keep_dim, int axis=*) except -1 nogil
    cdef int unpack_into(_StridedLayout self, _StridedLayout out_layout, int itemsize, int axis=*) except -1 nogil
    cdef int slice_into(_StridedLayout self, _StridedLayout out_layout, tuple slices) except -1

# ==============================
# Base layout helpers
# ==============================


cdef inline int init_base_layout(BaseLayout& layout, int ndim) except -1 nogil:
    if ndim > STRIDED_LAYOUT_MAX_NDIM:
        raise ValueError(f"Unsupported number of dimensions: {ndim}. Max supported ndim is {STRIDED_LAYOUT_MAX_NDIM}")
    # resize(0) is no op, that results in _mem.data() being NULL,
    # which would make it tricky to distinguish between strides == NULL
    # and strides == tuple()
    layout._mem.resize(2 * max(ndim, 1))
    layout.shape = layout._mem.data()
    layout.strides = layout._mem.data() + ndim
    layout.ndim = ndim
    return 0


cdef inline int trim_base_layout(BaseLayout& layout, int ndim) except -1 nogil:
    if ndim > layout.ndim:
        raise AssertionError(f"Cannot trim layout to {ndim} dimensions, it has {layout.ndim} dimensions")
    layout.ndim = ndim
    return 0


cdef inline void _swap_layout(BaseLayout& a, BaseLayout& b) noexcept nogil:
    _swap(a._mem, b._mem)
    _swap(a.shape, b.shape)
    _swap(a.strides, b.strides)
    _swap(a.ndim, b.ndim)


cdef inline void _assure_strides_ptr(BaseLayout& base) noexcept nogil:
    if base.strides == NULL:
        base.strides = base._mem.data() + base._mem.size() // 2


cdef inline stride_t *get_strides_ptr(BaseLayout& base) except? NULL nogil:
    if base.strides != NULL:
        return base.strides
    cdef stride_t* tmp_strides = base._mem.data() + base._mem.size() // 2
    _dense_strides_c_ptrs(base.ndim, base.shape, tmp_strides)
    return tmp_strides


cdef inline bint _base_layout_equal(BaseLayout& a, BaseLayout& b) noexcept nogil:
    if a.ndim != b.ndim:
        return False
    for i in range(a.ndim):
        if a.shape[i] != b.shape[i]:
            return False
    if a.strides != NULL or b.strides != NULL:
        if a.strides == NULL or b.strides == NULL:
            return False
        for i in range(a.ndim):
            if a.strides[i] != b.strides[i]:
                return False
    return True


@cython.overflowcheck(True)
cdef inline int64_t _volume(BaseLayout& base) except? -1 nogil:
    cdef int64_t vol = 1
    for i in range(base.ndim):
        vol *= base.shape[i]
    return vol


cdef inline int _divide_strides(BaseLayout& base, int itemsize) except -1 nogil:
    cdef stride_t stride
    if base.strides == NULL:
        raise ValueError("cannot divide strides, layout has no strides")
    for i in range(base.ndim):
        stride = base.strides[i] // itemsize
        if stride * itemsize != base.strides[i]:
            raise ValueError("strides must be divisible by itemsize")
        base.strides[i] = stride
    return 0


cdef inline void _zero_strides_ptr(int ndim, stride_t* strides) noexcept nogil:
    for i in range(ndim):
        strides[i] = 0


cdef inline void _zero_strides(BaseLayout& base) noexcept nogil:
    _assure_strides_ptr(base)
    _zero_strides_ptr(base.ndim, base.strides)


cdef inline stride_t _dense_strides_c_ptrs(int ndim, extent_t* shape, stride_t* strides) except? -1 nogil:
    cdef stride_t stride = 1
    cdef int i = ndim - 1
    while i >= 0:
        strides[i] = stride
        stride = _overflow_checked_mul(stride, shape[i])
        i -= 1
    if stride == 0:
        _zero_strides_ptr(ndim, strides)
    return stride


cdef inline stride_t _dense_strides_c(BaseLayout& base) except? -1 nogil:
    cdef int ndim = base.ndim
    _assure_strides_ptr(base)
    return _dense_strides_c_ptrs(ndim, base.shape, base.strides)


cdef inline stride_t _dense_strides_f(BaseLayout& base) except? -1 nogil:
    cdef int ndim = base.ndim
    _assure_strides_ptr(base)
    cdef stride_t stride = 1
    cdef int i = 0
    while i < ndim:
        base.strides[i] = stride
        stride = _overflow_checked_mul(stride, base.shape[i])
        i += 1
    if stride == 0:
        _zero_strides(base)
    return stride


cdef inline stride_t _dense_strides_in_order(BaseLayout& base, axis_vec_t& stride_order) except? -1 nogil:
    cdef int ndim = base.ndim
    if <size_t>ndim != stride_order.size():
        raise ValueError(f"stride_order must have the same length as shape. Shape has {ndim} dimensions, but stride_order has {stride_order.size()} elements.")
    _assure_strides_ptr(base)
    cdef stride_t stride = 1
    cdef int i = ndim - 1
    cdef axes_mask_t axis_order_mask = 0
    cdef axes_mask_t axis_mask
    cdef axis_t axis
    while i >= 0:
        axis = stride_order[i]
        if not _normalize_axis(axis, ndim):
            raise ValueError(f"Invalid stride order: axis {axis} out of range for {ndim}D tensor")
        axis_mask = _axis2mask(axis)
        if axis_order_mask & axis_mask:
            raise ValueError(f"The stride order must be a permutation. Axis {axis} appears multiple times.")
        axis_order_mask |= axis_mask
        base.strides[axis] = stride
        stride = _overflow_checked_mul(stride, base.shape[axis])
        i -= 1
    if stride == 0:
        _zero_strides(base)
    return stride


cdef inline bint _is_contiguous_c(int64_t volume, BaseLayout& base) except -1 nogil:
    if volume == 0 or base.strides == NULL:
        return True
    cdef int64_t stride = 1
    cdef int64_t j = base.ndim - 1
    cdef extent_t extent
    while j >= 0:
        extent = base.shape[j]
        if extent != 1:
            if base.strides[j] != stride:
                return False
            stride *= extent
        j -= 1
    return True


cdef inline bint _is_contiguous_f(int64_t volume, BaseLayout& base) except -1 nogil:
    if volume == 0:
        return True
    cdef int ndim = base.ndim
    cdef int64_t j = 0
    if base.strides == NULL:
        # find first non-singleton dimension
        while j < ndim and base.shape[j] == 1:
            j += 1
        # if any subsequent dimension is not a singleton, return False
        for i in range(j + 1, ndim):
            if base.shape[i] != 1:
                return False
        return True
    cdef int64_t stride = 1
    cdef extent_t extent
    while j < ndim:
        extent = base.shape[j]
        if extent != 1:
            if base.strides[j] != stride:
                return False
            stride *= extent
        j += 1
    return True


cdef inline bint _is_contiguous_any(int64_t volume, BaseLayout& base, axis_vec_t& axis_order) except -1 nogil:
    if volume == 0 or base.strides == NULL:
        return True
    cdef int64_t stride = 1
    cdef int64_t j = base.ndim - 1
    cdef axis_t axis
    cdef extent_t extent
    while j >= 0:
        axis = axis_order[j]
        extent = base.shape[axis]
        if extent != 1:
            if base.strides[axis] != stride:
                return False
            stride *= extent
        j -= 1
    return True


cdef inline int _validate_shape(BaseLayout& base) except -1 nogil:
    for i in range(base.ndim):
        if base.shape[i] < 0:
            raise ValueError("Extents must be non-negative")
    return 0


cdef inline int _init_base_layout_from_tuple(BaseLayout& base, tuple shape, tuple strides) except -1:
    cdef int ndim = len(shape)
    init_base_layout(base, ndim)
    for i in range(ndim):
        base.shape[i] = shape[i]
    _validate_shape(base)

    if strides is None:
        base.strides = NULL
    else:
        if len(strides) != ndim:
            raise ValueError(f"Strides, if provided, must have the same length as shape. Shape has {ndim} dimensions, but strides has {len(strides)} elements.")
        for i in range(ndim):
            base.strides[i] = strides[i]
    return 0


cdef inline int _init_base_layout_from_ptr(BaseLayout& base, int ndim, extent_t* shape, stride_t* strides) except -1 nogil:
    init_base_layout(base, ndim)
    for i in range(ndim):
        base.shape[i] = shape[i]
    _validate_shape(base)

    if strides == NULL:
        base.strides = NULL
    else:
        for i in range(ndim):
            base.strides[i] = strides[i]
    return 0

# ==============================
# Strided layout helpers
# ==============================


cdef inline bint _has_valid_property(_StridedLayout self, Property prop) noexcept nogil:
    return self._prop_mask & prop


cdef inline void _mark_property_valid(_StridedLayout self, Property prop) noexcept nogil:
    self._prop_mask |= prop


cdef inline bint _boolean_property(_StridedLayout self, Property prop) noexcept nogil:
    return self._boolean_props & prop


cdef inline bint _set_boolean_property(_StridedLayout self, Property prop, bint value) noexcept nogil:
    if value:
        self._boolean_props |= prop
    else:
        self._boolean_props &= ~prop
    _mark_property_valid(self, prop)
    return value


# ==============================
# Conversion, validation and normalization helpers
# ==============================


cdef inline axes_mask_t _axis2mask(axis_t axis) noexcept nogil:
    return 1ULL << axis


cdef inline axes_mask_t flatten_all_axes_mask(int ndim) noexcept nogil:
    if ndim == 0:
        return 0
    elif ndim == STRIDED_LAYOUT_MAX_NDIM:
        return AXES_MASK_ALL
    return (1ULL << ndim) - 2


cdef inline OrderFlag _stride_order2vec(axis_vec_t& stride_order_vec, object stride_order) except? ORDER_NONE:
    if stride_order == 'C':
        return ORDER_C
    elif stride_order == 'F':
        return ORDER_F
    elif isinstance(stride_order, tuple | list):
        _tuple2axis_vec(stride_order_vec, stride_order)
        return ORDER_PERM
    return ORDER_NONE


cdef inline int _tuple2axis_vec(axis_vec_t& vec, object t) except -1:
    cdef int ndim = len(t)
    vec.resize(ndim)
    for i in range(ndim):
        vec[i] = t[i]
    return 0


cdef inline bint _normalize_axis(integer_t& axis, integer_t extent) except -1 nogil:
    if axis < -extent or axis >= extent:
        return False
    if axis < 0:
        axis += extent
    return True


cdef inline bint _is_unique(BaseLayout& base, axis_vec_t& stride_order) except -1 nogil:
    if base.strides == NULL:
        return True
    cdef int64_t cur_max_offset = 0
    cdef int i = base.ndim - 1
    cdef int64_t stride
    cdef axis_t axis
    cdef extent_t extent
    while i >= 0:
        axis = stride_order[i]
        extent = base.shape[axis]
        if extent != 1:
            stride = _c_abs(base.strides[axis])
            if cur_max_offset >= stride:
                return False
            cur_max_offset = _overflow_checked_sum(cur_max_offset, _overflow_checked_mul(stride, (extent - 1)))
        i -= 1
    return True


@cython.overflowcheck(True)
cdef inline int64_t _overflow_checked_mul(int64_t a, int64_t b) except? -1 nogil:
    return a * b


@cython.overflowcheck(True)
cdef inline int64_t _overflow_checked_diff(int64_t a, int64_t b) except? -1 nogil:
    return a - b


@cython.overflowcheck(True)
cdef inline int64_t _overflow_checked_sum(int64_t a, int64_t b) except? -1 nogil:
    return a + b


@cython.overflowcheck(True)
cdef inline int64_t _overflow_checked_div_ceil(int64_t a, int64_t b) except? -1 nogil:
    return (a + b - 1) // b



# SPDX-FileCopyrightText: Copyright (c) 2024-2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
#
# SPDX-License-Identifier: Apache-2.0

from ._buffer import *
from ._device_memory_resource import *
from ._graph_memory_resource import *
from ._ipc import *
from ._legacy import *
from ._managed_memory_resource import *
from ._pinned_memory_resource import *
from ._virtual_memory_resource import *



# SPDX-FileCopyrightText: Copyright (c) 2025-2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
#
# SPDX-License-Identifier: Apache-2.0

from cuda.core.graph._graph_builder import (
    Graph,
    GraphBuilder,
    GraphCompleteOptions,
    GraphDebugPrintOptions,
)
from cuda.core.graph._graph_def import (
    Condition,
    GraphAllocOptions,
    GraphDef,
)
from cuda.core.graph._graph_node import GraphNode
from cuda.core.graph._subclasses import (
    AllocNode,
    ChildGraphNode,
    ConditionalNode,
    EmptyNode,
    EventRecordNode,
    EventWaitNode,
    FreeNode,
    HostCallbackNode,
    IfElseNode,
    IfNode,
    KernelNode,
    MemcpyNode,
    MemsetNode,
    SwitchNode,
    WhileNode,
)



# SPDX-FileCopyrightText: Copyright (c) 2024-2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
#
# SPDX-License-Identifier: Apache-2.0

cimport cpython
from cpython.object cimport PyObject
from libc.stdint cimport int64_t, int32_t, uint8_t, uint16_t, uint32_t

from cuda.bindings cimport cydriver, cynvrtc, cynvvm, cynvjitlink


ctypedef fused integer_t:
    int64_t
    int32_t


# mimic CU_DEVICE_INVALID
cdef const cydriver.CUcontext CU_CONTEXT_INVALID = <cydriver.CUcontext>(-2)


cdef int HANDLE_RETURN(cydriver.CUresult err) except?-1 nogil
cdef int HANDLE_RETURN_NVRTC(cynvrtc.nvrtcProgram prog, cynvrtc.nvrtcResult err) except?-1 nogil
cdef int HANDLE_RETURN_NVVM(cynvvm.nvvmProgram prog, cynvvm.nvvmResult err) except?-1 nogil
cdef int HANDLE_RETURN_NVJITLINK(
    cynvjitlink.nvJitLinkHandle handle, cynvjitlink.nvJitLinkResult err) except?-1 nogil


# TODO: stop exposing these within the codebase?
cpdef int _check_driver_error(cydriver.CUresult error) except?-1 nogil
cpdef int _check_runtime_error(error) except?-1
cpdef int _check_nvrtc_error(error) except?-1


cpdef check_or_create_options(type cls, options, str options_description=*, bint keep_none=*)

cpdef tuple _parse_fill_value(value)


# Create low-level externs so Cython won't "helpfully" handle reference counting
# for us.  Prefixing with an underscore to distinguish it from the definition in
# cpython.long.
cdef extern from "Python.h":
    PyObject *_PyLong_FromLongLong "PyLong_FromLongLong" (long long val) except NULL
    void _PyTuple_SET_ITEM "PyTuple_SET_ITEM" (object p, Py_ssize_t pos, PyObject *o)


cdef inline tuple carray_integer_t_to_tuple(integer_t *ptr, int length):
    # Construct shape and strides tuples using the Python/C API for speed
    cdef tuple result = cpython.PyTuple_New(length)
    for i in range(length):
        _PyTuple_SET_ITEM(result, i, _PyLong_FromLongLong(ptr[i]))
    return result



# SPDX-FileCopyrightText: Copyright (c) 2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
#
# SPDX-License-Identifier: Apache-2.0

from cuda.bindings cimport cydriver


cdef bint _is_py_host_trampoline(cydriver.CUhostFn fn) noexcept nogil

cdef void _attach_user_object(
    cydriver.CUgraph graph, void* ptr,
    cydriver.CUhostFn destroy) except *

cdef void _attach_host_callback_to_graph(
    cydriver.CUgraph graph, object fn, object user_data,
    cydriver.CUhostFn* out_fn, void** out_user_data) except *



# SPDX-FileCopyrightText: Copyright (c) 2025-2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: LicenseRef-NVIDIA-SOFTWARE-LICENSE

"""Internal support for error-enum explanations.

``cuda_core`` keeps frozen 13.1.1 fallback tables for older ``cuda-bindings``
releases. Driver/runtime error enums carry usable ``__doc__`` text starting in
the 12.x backport line at ``cuda-bindings`` 12.9.6, and in the mainline 13.x
series at ``cuda-bindings`` 13.2.0. This module decides which source to use
and normalizes generated docstrings so user-facing ``CUDAError`` messages stay
presentable.

The cleanup rules here were derived while validating generated enum docstrings
in PR #1805. Keep them narrow and remove them when codegen quirks or fallback
support are no longer needed.
"""

from __future__ import annotations

import importlib.metadata
import re
from collections.abc import Callable
from typing import Any

_MIN_12X_BINDING_VERSION_FOR_ENUM_DOCSTRINGS = (12, 9, 6)
_MIN_13X_BINDING_VERSION_FOR_ENUM_DOCSTRINGS = (13, 2, 0)
_RST_INLINE_ROLE_RE = re.compile(r":(?:[a-z]+:)?[a-z]+:`([^`]+)`")
_WORDWRAP_HYPHEN_AFTER_RE = re.compile(r"(?<=[0-9A-Za-z_])- (?=[0-9A-Za-z_])")
_WORDWRAP_HYPHEN_BEFORE_RE = re.compile(r"(?<=[0-9A-Za-z_]) -(?=[0-9A-Za-z_])")
_ExplanationTable = dict[int, str | tuple[str, ...]]
_ExplanationTableLoader = Callable[[], _ExplanationTable]


# ``version.pyx`` cannot be reused here (circular import via ``cuda_utils``).
def _binding_version() -> tuple[int, int, int]:
    """Return the installed ``cuda-bindings`` version, or a conservative old value."""
    try:
        parts = importlib.metadata.version("cuda-bindings").split(".")[:3]
    except importlib.metadata.PackageNotFoundError:
        return (0, 0, 0)  # For very old versions of cuda-python
    return tuple(int(v) for v in parts)


def _binding_version_has_usable_enum_docstrings(version: tuple[int, int, int]) -> bool:
    """Whether released bindings are known to carry usable error-enum ``__doc__`` text."""
    return (
        _MIN_12X_BINDING_VERSION_FOR_ENUM_DOCSTRINGS <= version < (13, 0, 0)
        or version >= _MIN_13X_BINDING_VERSION_FOR_ENUM_DOCSTRINGS
    )


def _fix_hyphenation_wordwrap_spacing(s: str) -> str:
    """Remove spaces around hyphens introduced by line wrapping in generated ``__doc__`` text.

    This targets asymmetric wrap artifacts such as ``non- linear`` or
    ``GPU- Direct`` while leaving intentional ``a - b`` separators alone.
    """
    prev = None
    while prev != s:
        prev = s
        s = _WORDWRAP_HYPHEN_AFTER_RE.sub("-", s)
        s = _WORDWRAP_HYPHEN_BEFORE_RE.sub("-", s)
    return s


def clean_enum_member_docstring(doc: str | None) -> str | None:
    """Turn an enum member ``__doc__`` into plain text.

    The generated enum docstrings are already close to user-facing prose, but
    they may contain Sphinx inline roles, line wrapping, or a small known
    codegen defect. Normalize only those differences so the text is suitable
    for error messages.
    """
    if doc is None:
        return None
    s = doc
    # Known codegen bug on cudaErrorIncompatibleDriverContext. Remove once fixed
    # in cuda-bindings code generation.
    s = s.replace("\n:py:obj:`~.Interactions`", ' "Interactions ')
    # Drop a leading "~." or "." after removing the surrounding RST inline role.
    s = _RST_INLINE_ROLE_RE.sub(lambda m: re.sub(r"^~?\.", "", m.group(1)), s)
    # Strip simple bold emphasis markers.
    s = re.sub(r"\*\*([^*]+)\*\*", r"\1", s)
    # Strip simple italic emphasis markers.
    s = re.sub(r"\*([^*]+)\*", r"\1", s)
    # Collapse wrapped lines and repeated spaces.
    s = re.sub(r"\s+", " ", s).strip()
    s = _fix_hyphenation_wordwrap_spacing(s)
    return s


class DocstringBackedExplanations:
    """Compatibility shim exposing enum-member ``__doc__`` text via ``dict.get``.

    Keeps the existing ``.get(int(error))`` lookup shape used by ``cuda_utils.pyx``.
    """

    __slots__ = ("_enum_type",)

    def __init__(self, enum_type: Any) -> None:
        self._enum_type = enum_type

    def get(self, code: int, default: str | None = None) -> str | None:
        try:
            member = self._enum_type(code)
        except ValueError:
            return default

        raw_doc = member.__doc__
        if raw_doc is None:
            return default

        return clean_enum_member_docstring(raw_doc)


def get_best_available_explanations(
    enum_type: Any,
    fallback: _ExplanationTable | _ExplanationTableLoader,
) -> DocstringBackedExplanations | _ExplanationTable:
    """Pick one explanation source per bindings version.

    Use enum-member ``__doc__`` only for bindings versions known to expose
    usable per-member text (12.9.6+ in the 12.x backport line, 13.2.0+ in the
    13.x mainline). Otherwise keep using the frozen 13.1.1 fallback tables.
    """
    if not _binding_version_has_usable_enum_docstrings(_binding_version()):
        if callable(fallback):
            return fallback()
        return fallback
    return DocstringBackedExplanations(enum_type)



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
#
# SPDX-License-Identifier: Apache-2.0

from cuda.core._resource_handles cimport ContextHandle, StreamHandle


cdef class Stream:

    cdef:
        StreamHandle _h_stream
        ContextHandle _h_context
        int _device_id
        int _nonblocking
        int _priority
        object __weakref__

    @staticmethod
    cdef Stream _from_handle(type cls, StreamHandle h_stream)

    cpdef close(self)


cpdef Stream default_stream()
cdef Stream Stream_accept(arg, bint allow_stream_protocol=*)



# SPDX-FileCopyrightText: Copyright (c) 2024-2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
#
# SPDX-License-Identifier: Apache-2.0

from cuda.core._memory._memory_pool cimport _MemPool
from cuda.core._memory._ipc cimport IPCDataForMR


cdef class DeviceMemoryResource(_MemPool):
    cdef:
        int _dev_id
        object _peer_accessible_by


cpdef DMR_mempool_get_access(DeviceMemoryResource, int)



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
#
# SPDX-License-Identifier: LicenseRef-NVIDIA-SOFTWARE-LICENSE


def assert_type(obj, expected_type):
    """Ensure obj is of expected_type, else raise AssertionError with a clear message."""
    if not isinstance(obj, expected_type):
        raise TypeError(f"Expected type {expected_type.__name__}, but got {type(obj).__name__}")


def assert_type_str_or_bytes_like(obj):
    """Ensure obj is of type str or bytes, else raise AssertionError with a clear message."""
    if not isinstance(obj, (str, bytes, bytearray)):
        raise TypeError(f"Expected type str or bytes or bytearray, but got {type(obj).__name__}")


def raise_code_path_meant_to_be_unreachable():
    raise RuntimeError("This code path is meant to be unreachable.")



# SPDX-FileCopyrightText: Copyright (c) 2024-2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
#
# SPDX-License-Identifier: Apache-2.0

from cuda.core._memory._buffer cimport MemoryResource


cdef class cyGraphMemoryResource(MemoryResource):
    cdef:
        int _device_id



# SPDX-FileCopyrightText: Copyright (c) 2024-2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
#
# SPDX-License-Identifier: Apache-2.0

from cuda.bindings cimport cydriver
from cuda.core._memory._buffer cimport MemoryResource
from cuda.core._memory._ipc cimport IPCDataForMR
from cuda.core._resource_handles cimport MemoryPoolHandle


cdef class _MemPool(MemoryResource):
    cdef:
        MemoryPoolHandle      _h_pool
        bint                  _mempool_owned
        IPCDataForMR          _ipc_data
        object                _attributes
        object                __weakref__


cdef int MP_init_create_pool(
    _MemPool self,
    cydriver.CUmemLocationType loc_type,
    int loc_id,
    cydriver.CUmemAllocationType alloc_type,
    bint ipc_enabled,
    size_t max_size,
) except? -1

cdef int MP_init_current_pool(
    _MemPool self,
    cydriver.CUmemLocationType loc_type,
    int loc_id,
    cydriver.CUmemAllocationType alloc_type,
) except? -1

cdef int MP_raise_release_threshold(_MemPool self) except? -1


cdef class _MemPoolAttributes:
    cdef:
        MemoryPoolHandle _h_pool

    @staticmethod
    cdef _MemPoolAttributes _init(MemoryPoolHandle h_pool)

    cdef int _getattribute(self, cydriver.CUmemPool_attribute attr_enum, void* value) except? -1
